import argparse
import os
import secrets
import string
import subprocess
from pathlib import Path

import yaml

from smartem.data_model import setup


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--data",
        help="Directory where database wil be initialised",
        dest="data_dir",
    )
    parser.add_argument(
        "--port",
        help="Port the database server will serve, default is 5432",
        dest="port",
        default=5432,
    )
    parser.add_argument(
        "--make-credentials-file",
        help="Make a credentials file for the database for use by smartem, default is True",
        dest="make_creds",
        default=True,
    )
    args = parser.parse_args()

    alphabet = string.ascii_letters + string.digits
    pswd_su = "".join(secrets.choice(alphabet) for i in range(20))

    with open(".postgres_pwd.txt", "w") as pwdfile:
        pwdfile.write(pswd_su)

    initdb_cmd = [
        "initdb",
        "-A",
        "md5",
        "-D",
        args.data_dir,
        "--pwfile=.postgres_pwd.txt",
    ]
    initdb = subprocess.run(initdb_cmd)
    Path(".postgres_pwd.txt").unlink()
    if initdb.returncode:
        exit(f"Database initialisation failed with return code {initdb.returncode}")
    else:
        print("Database initialisation complete")

        os.environ["PGPASSWORD"] = pswd_su

        server_start = subprocess.run(
            [
                "pg_ctl",
                "-o",
                f'"-p {args.port}"',
                "-D",
                args.data_dir,
                "-l" f"{Path(args.data_dir) / 'logfile'}",
                "start",
            ]
        )
        if server_start.returncode:
            exit(
                f"FATAL: Database server failed to start with return code {server_start.returncode}"
            )

        if args.make_creds:
            creds = {
                "username": os.getenv("USER"),
                "password": pswd_su,
                "host": "localhost",
                "port": args.port,
                "database": "smartem",
            }
            with open(Path(args.data_dir) / "credentials.yaml", "w") as credfile:
                yaml.dump(creds, credfile)

        create_db = subprocess.run(["createdb", "smartem"])
        if create_db.returncode:
            exit(
                f"FATAL: smartem database was not created with return code {server_start.returncode}"
            )

        os.environ["SMARTEM_CREDENTIALS"] = str(
            (Path(args.data_dir) / "credentials.yaml").resolve()
        )
        setup()

        subprocess.run(["pg_ctl", "-D", args.data_dir, "stop"])

        print(
            f"Run the following to setup database access credentials:\n export SMARTEM_CREDENTIALS={(Path(args.data_dir) / 'credentials.yaml').resolve()}"
        )



from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache
from pathlib import Path
from typing import Dict, Generator, List, Optional

from PyQt5.QtWidgets import (
    QComboBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QVBoxLayout,
)

from smartem.data_model import ParticleSet, ParticleSetLinker
from smartem.data_model.extract import DataAPI
from smartem.gui.qt.component_tab import ComponentTab, background
from smartem.parsing.star import (
    get_column_data,
    get_columns,
    insert_exposure_data,
    insert_particle_data,
    insert_particle_set,
    open_star_file,
)


@lru_cache(maxsize=5)
def relevant_star_files(directory: Path) -> List[str]:
    sfiles = []
    for sf in directory.glob("*/*/*.star"):
        str_sf = str(sf)
        if (
            all(p not in str_sf for p in ("gui", "pipeline", "Nodes", "NODES"))
            and "job" not in sf.name
            and not sf.parent.is_symlink()
        ):
            sfiles.append(str_sf)
    return sfiles


def _string_to_glob(glob_string: str) -> Generator[Path, None, None]:
    split_string = glob_string.split("/")
    if "*" in split_string[0]:
        return Path("/").glob("/".join(split_string)[1:])
    root_path = Path("/")
    end_index = 0
    for i, s in enumerate(split_string):
        if "*" not in s:
            root_path = root_path / s
        else:
            end_index = i
            break
    return root_path.glob("/".join(split_string[end_index:]))


class StarDataLoader(ComponentTab):
    def __init__(
        self,
        extractor: DataAPI,
        project_directory: Optional[Path] = None,
        refreshers: Optional[List[ComponentTab]] = None,
    ):
        super().__init__(refreshers=refreshers)
        self._extractor = extractor
        self.grid = QGridLayout()
        self.setLayout(self.grid)
        self._exposure_tag = None
        self._column = None
        self._proj_dir = project_directory
        self.project = ""

        star_lbl = QLabel()
        star_lbl.setText("Star file:")
        column_lbl = QLabel()
        column_lbl.setText("Data column:")

        self._file_combo = QComboBox()
        self._file_combo.setEditable(True)
        self._file_combo.currentIndexChanged.connect(self._select_star_file)

        star_hbox = QHBoxLayout()
        star_hbox.addWidget(star_lbl, 1)
        star_hbox.addWidget(self._file_combo, 1)

        self._file_vbox = QVBoxLayout()
        self._file_vbox.addLayout(star_hbox, 1)

        self.grid.addLayout(self._file_vbox, 2, 1, 1, 2)
        self._column_combo = QComboBox()
        self._column_combo.setEditable(True)
        self._column_combo.currentIndexChanged.connect(self._select_column)

        column_hbox = QHBoxLayout()
        column_hbox.addWidget(column_lbl, 1)
        column_hbox.addWidget(self._column_combo, 1)

        self.grid.addLayout(column_hbox, 3, 1, 1, 2)

        load_btn = QPushButton("Load")
        load_btn.clicked.connect(self.load)
        self.grid.addWidget(load_btn, 5, 1)

    @background(children=None)
    def _set_project_directory(self, project_directory: Path):
        self._proj_dir = project_directory
        star_files = relevant_star_files(project_directory)
        self._file_combo.addItems(star_files)

    def _select_star_file(
        self,
        index: int,
        column_combos: Optional[List[QComboBox]] = None,
        defaults: Optional[List[str]] = None,
        connections: Optional[Dict[str, str]] = None,
    ):
        if "*" in self._file_combo.currentText():
            star_file_path = next(_string_to_glob(self._file_combo.currentText()))
        else:
            star_file_path = Path(self._file_combo.currentText())
        try:
            star_file = open_star_file(star_file_path)
        except (OSError, ValueError):
            print(f"Could not open star file {star_file_path}")
            return
        if not column_combos:
            column_combos = [self._column_combo]
        columns = get_columns(star_file, ignore=["pipeline"])
        for combo in column_combos:
            combo.clear()
            combo.addItem("")
        for c in sorted(set(columns)):
            for i, combo in enumerate(column_combos):
                combo.addItem(c)
                if defaults and c == defaults[i]:
                    combo.setCurrentText(c)
                    if connections and connections.get(defaults[i]):
                        setattr(self, connections[defaults[i]], c)

    def _select_column(self, index: int):
        self._column = self._column_combo.currentText()

    def load(self, **kwargs):
        raise NotImplementedError


class ExposureDataLoader(StarDataLoader):
    def __init__(
        self,
        extractor: DataAPI,
        project_directory: Optional[Path] = None,
        refreshers: Optional[List[ComponentTab]] = None,
    ):
        super().__init__(extractor, project_directory, refreshers=refreshers)
        exposure_lbl = QLabel()
        exposure_lbl.setText("Micrograph identifier:")

        self._exposure_tag_combo = QComboBox()
        self._exposure_tag_combo.setEditable(True)
        self._exposure_tag_combo.currentIndexChanged.connect(self._select_exposure_tag)

        self._identifier_box = QVBoxLayout()

        hbox = QHBoxLayout()
        hbox.addWidget(exposure_lbl, 1)
        hbox.addWidget(self._exposure_tag_combo, 1)

        self._identifier_box.addLayout(hbox, 1)

        self.grid.addLayout(self._identifier_box, 4, 1, 1, 2)

        self._exposure_tag = self._exposure_tag_combo.currentText()

    def _select_exposure_tag(self, index: int):
        self._exposure_tag = self._exposure_tag_combo.currentText()

    def _select_star_file(
        self,
        index: int,
        column_combos: Optional[List[QComboBox]] = None,
        defaults: Optional[List[str]] = None,
        connections: Optional[Dict[str, str]] = None,
    ):
        super()._select_star_file(
            index,
            column_combos=column_combos
            or [self._column_combo, self._exposure_tag_combo],
            defaults=defaults or ["", "_rlnmicrographname"],
            connections=connections or {"_rlnmicrographname": "_exposure_tag"},
        )

    def _insert_from_star_file(self, star_file_path: Path):
        if self._exposure_tag:
            star_file = open_star_file(star_file_path)
            column_data = get_column_data(
                star_file, [self._exposure_tag, self._column], "micrographs"
            )
            insert_exposure_data(
                column_data,
                self._exposure_tag,
                str(star_file_path),
                self._extractor,
                project=self.project,
            )

    def load(self):
        if (
            self._exposure_tag or self._exposure_tag_combo.currentText()
        ) and self._column:
            if not self._exposure_tag:
                self._exposure_tag = self._exposure_tag_combo.currentText()
            if "*" in self._file_combo.currentText():
                for sfp in _string_to_glob(self._file_combo.currentText()):
                    if not sfp.parent.is_symlink():
                        self._insert_from_star_file(Path(sfp))
            else:
                self._insert_from_star_file(Path(self._file_combo.currentText()))
            self.refresh()


class ParticleDataLoader(ExposureDataLoader):
    def __init__(
        self,
        extractor: DataAPI,
        project_directory: Optional[Path] = None,
        refreshers: Optional[List[ComponentTab]] = None,
    ):
        super().__init__(extractor, project_directory, refreshers=refreshers)

        x_lbl = QLabel()
        x_lbl.setText("x coordinate identifier:")

        self._x_tag_combo = QComboBox()
        self._x_tag_combo.setEditable(True)
        self._x_tag_combo.currentIndexChanged.connect(self._select_x_tag)

        x_hbox = QHBoxLayout()
        x_hbox.addWidget(x_lbl, 1)
        x_hbox.addWidget(self._x_tag_combo, 1)

        self._identifier_box.addLayout(x_hbox, 1)

        y_lbl = QLabel()
        y_lbl.setText("y coordinate identifier:")

        self._y_tag_combo = QComboBox()
        self._y_tag_combo.setEditable(True)
        self._y_tag_combo.currentIndexChanged.connect(self._select_y_tag)

        y_hbox = QHBoxLayout()
        y_hbox.addWidget(y_lbl, 1)
        y_hbox.addWidget(self._y_tag_combo, 1)

        self._identifier_box.addLayout(y_hbox, 1)

    def _select_x_tag(self, index: int):
        self._x_tag = self._x_tag_combo.currentText()

    def _select_y_tag(self, index: int):
        self._y_tag = self._y_tag_combo.currentText()

    def _select_star_file(
        self,
        index: int,
        column_combos: Optional[List[QComboBox]] = None,
        defaults: Optional[List[str]] = None,
        connections: Optional[Dict[str, str]] = None,
    ):
        super()._select_star_file(
            index,
            column_combos=column_combos
            or [
                self._column_combo,
                self._exposure_tag_combo,
                self._x_tag_combo,
                self._y_tag_combo,
            ],
            defaults=defaults
            or ["", "_rlnmicrographname", "_rlncoordinatex", "_rlncoordinatey"],
            connections=connections
            or {
                "_rlnmicrographname": "_exposure_tag",
                "_rlncoordinatex": "_x_tag",
                "_rlncoordinatey": "_y_tag",
            },
        )

    def _insert_from_star_file(
        self, star_file_path: Path, just_particles: bool = False
    ):
        star_file = open_star_file(star_file_path)
        if self._exposure_tag:
            if just_particles:
                column_data = get_column_data(
                    star_file,
                    [self._exposure_tag, self._x_tag, self._y_tag],
                    "particles",
                )
                particles = insert_particle_data(
                    column_data,
                    self._exposure_tag,
                    self._x_tag,
                    self._y_tag,
                    str(star_file_path),
                    self._extractor,
                    project=self.project,
                )
                source_set = ParticleSet(
                    group_name=str(star_file_path),
                    identifier=str(star_file_path),
                    project_name=self._extractor._project,
                )
                self._extractor.put([source_set])
                linkers = [
                    ParticleSetLinker(
                        set_name=str(star_file_path), particle_id=p.particle_id
                    )
                    for p in particles
                ]
                self._extractor.put(linkers)

            else:
                column_data = get_column_data(
                    star_file,
                    [self._exposure_tag, self._x_tag, self._y_tag, self._column],
                    "particles",
                )
                insert_particle_data(
                    column_data,
                    self._exposure_tag,
                    self._x_tag,
                    self._y_tag,
                    str(star_file_path),
                    self._extractor,
                    project=self.project,
                )

    def load(self):
        if self._exposure_tag and self._x_tag and self._y_tag and self._column:
            if "*" in self._file_combo.currentText():
                for sfp in _string_to_glob(self._file_combo.currentText()):
                    if not sfp.parent.is_symlink():
                        self._insert_from_star_file(Path(sfp))
            else:
                self._insert_from_star_file(Path(self._file_combo.currentText()))
        elif self._exposure_tag and self._x_tag and self._y_tag:
            if "*" in self._file_combo.currentText():
                for sfp in _string_to_glob(self._file_combo.currentText()):
                    if not sfp.parent.is_symlink():
                        self._insert_from_star_file(Path(sfp), just_particles=True)
            else:
                self._insert_from_star_file(
                    Path(self._file_combo.currentText()), just_particles=True
                )
        self.refresh()


class ParticleSetDataLoader(ParticleDataLoader):
    def __init__(
        self,
        extractor: DataAPI,
        project_directory: Optional[Path] = None,
        refreshers: Optional[List[ComponentTab]] = None,
    ):
        super().__init__(extractor, project_directory, refreshers=refreshers)

        self._group_name_box = QLineEdit()
        self.grid.addWidget(self._group_name_box, 5, 2)

        set_id_lbl = QLabel()
        set_id_lbl.setText("Particle set identifier:")

        self._set_id_combo = QComboBox()
        self._set_id_combo.setEditable(True)
        self._set_id_combo.currentIndexChanged.connect(self._select_set_id_tag)

        set_id_hbox = QHBoxLayout()
        set_id_hbox.addWidget(set_id_lbl, 1)
        set_id_hbox.addWidget(self._set_id_combo, 1)

        self._identifier_box.addLayout(set_id_hbox, 1)

        cross_ref_lbl = QLabel()
        cross_ref_lbl.setText("Cross reference column:")

        self._cross_ref_combo = QComboBox()
        self._cross_ref_combo.setEditable(True)
        self._cross_ref_combo.setEnabled(False)

        cross_ref_hbox = QHBoxLayout()
        cross_ref_hbox.addWidget(cross_ref_lbl, 1)
        cross_ref_hbox.addWidget(self._cross_ref_combo, 1)

        self._identifier_box.addLayout(cross_ref_hbox, 1)

        cross_ref_file_hbox = QHBoxLayout()

        cross_ref_file_lbl = QLabel()
        cross_ref_file_lbl.setText("Star file for cross reference:")
        cross_ref_file_hbox.addWidget(cross_ref_file_lbl, 1)

        self._cross_ref_file_combo = QComboBox()
        self._cross_ref_file_combo.setEditable(True)
        self._cross_ref_file_combo.currentIndexChanged.connect(
            self._select_cross_ref_file
        )

        cross_ref_file_hbox.addWidget(self._cross_ref_file_combo, 1)

        self._file_vbox.addLayout(cross_ref_file_hbox, 1)

    @background(children=None)
    def _set_project_directory(self, project_directory: Path):
        self._proj_dir = project_directory
        self._file_combo.clear()
        self._cross_ref_file_combo.clear()
        self._cross_ref_file_combo.addItem("")
        for sf in self._proj_dir.glob("*/*/*.star"):
            str_sf = str(sf)
            if (
                all(p not in str_sf for p in ("gui", "pipeline", "Nodes", "NODES"))
                and "job" not in sf.name
                and not sf.parent.is_symlink()
            ):
                self._file_combo.addItem(str_sf)
                self._cross_ref_file_combo.addItem(str_sf)

    def _select_cross_ref_file(self):
        if self._cross_ref_file_combo.currentText():
            self._cross_ref_combo.setEnabled(True)
            star_file_path = Path(self._cross_ref_file_combo.currentText())
            try:
                star_file = open_star_file(star_file_path)
            except (OSError, ValueError):
                print(f"Could not open star file {star_file_path}")
                return
            columns = get_columns(star_file, ignore=["pipeline"])
            self._cross_ref_combo.clear()
            self._column_combo.clear()
            for c in sorted(set(columns)):
                self._cross_ref_combo.addItem(c)
                if c == "_rlnreferenceimage":
                    self._cross_ref_combo.setCurrentText(c)
                self._column_combo.addItem(c)

    def _select_set_id_tag(self, index: int):
        self._set_id_tag = self._set_id_combo.currentText()

    def _select_star_file(
        self,
        index: int,
        column_combos: Optional[List[QComboBox]] = None,
        defaults: Optional[List[str]] = None,
        connections: Optional[Dict[str, str]] = None,
    ):
        if self._cross_ref_file_combo.currentText():
            self._cross_ref_combo.clear()
            super()._select_star_file(
                index,
                column_combos=column_combos
                or [
                    self._exposure_tag_combo,
                    self._x_tag_combo,
                    self._y_tag_combo,
                    self._set_id_combo,
                ],
                defaults=defaults
                or [
                    "_rlnmicrographname",
                    "_rlncoordinatex",
                    "_rlncoordinatey",
                    "_rlnclassnumber",
                ],
                connections=connections
                or {
                    "_rlnmicrographname": "_exposure_tag",
                    "_rlncoordinatex": "_x_tag",
                    "_rlncoordinatey": "_y_tag",
                    "_rlnclassnumber": "_set_id_tag",
                },
            )
            star_file_path = Path(self._cross_ref_file_combo.currentText())
            try:
                star_file = open_star_file(star_file_path)
            except (OSError, ValueError):
                return
            columns = get_columns(star_file, ignore=["pipeline"])
            self._cross_ref_combo.clear()
            self._column_combo.clear()
            for c in sorted(set(columns)):
                self._cross_ref_combo.addItem(c)
                if c == "_rlnreferenceimage":
                    self._cross_ref_combo.setCurrentText(c)
                self._column_combo.addItem(c)
        else:
            super()._select_star_file(
                index,
                column_combos=column_combos
                or [
                    self._column_combo,
                    self._exposure_tag_combo,
                    self._x_tag_combo,
                    self._y_tag_combo,
                    self._set_id_combo,
                ],
                defaults=defaults
                or [
                    "",
                    "_rlnmicrographname",
                    "_rlncoordinatex",
                    "_rlncoordinatey",
                    "_rlnclassnumber",
                ],
            )

    def _insert_from_star_file(
        self,
        star_file_path: Path,
        just_particles: bool = False,
        cross_ref_file_path: Optional[Path] = None,
        new_thread: bool = True,
    ):
        print("insert from", star_file_path)
        if new_thread:
            data_api: Optional[DataAPI] = DataAPI()
        else:
            data_api = None
        if self._exposure_tag and self._column:
            if cross_ref_file_path:
                star_file = open_star_file(star_file_path)
                column_data = get_column_data(
                    star_file,
                    [
                        self._exposure_tag,
                        self._x_tag,
                        self._y_tag,
                        self._set_id_tag,
                        # self._cross_ref_combo.currentText(),
                    ],
                    "particles",
                )
                cross_ref_file = open_star_file(cross_ref_file_path)
                cross_ref_column_data = get_column_data(
                    cross_ref_file,
                    [self._cross_ref_combo.currentText(), self._column],
                    "model_classes",
                )
                for v in cross_ref_column_data.values():
                    for i, elem in enumerate(v):
                        if isinstance(elem, str) and "@" in elem:
                            _elem = elem.split("@")[0]
                            while _elem.startswith("0"):
                                _elem = _elem[1:]
                            v[i] = _elem
                cross_ref_dict = {}
                for i, k02 in enumerate(column_data[self._set_id_tag]):
                    for j, k01 in enumerate(
                        cross_ref_column_data[self._cross_ref_combo.currentText()]
                    ):
                        if k01 == str(k02):
                            cross_ref_dict[i] = cross_ref_column_data[self._column][j]
                column_data[self._column] = [
                    cross_ref_dict[i] for i in range(len(column_data[self._set_id_tag]))
                ]
                insert_particle_set(
                    column_data,
                    self._group_name_box.text(),
                    self._set_id_tag,
                    self._exposure_tag,
                    self._x_tag,
                    self._y_tag,
                    str(star_file_path),
                    data_api or self._extractor,
                    self.project,
                    add_source_to_id=True,
                )
            else:
                star_file = open_star_file(star_file_path)
                column_data = get_column_data(
                    star_file,
                    [
                        self._exposure_tag,
                        self._x_tag,
                        self._y_tag,
                        self._column,
                        self._set_id_tag,
                    ],
                    "particles",
                )
                insert_particle_set(
                    column_data,
                    self._group_name_box.text(),
                    self._set_id_tag,
                    self._exposure_tag,
                    self._x_tag,
                    self._y_tag,
                    str(star_file_path),
                    data_api or self._extractor,
                    self.project,
                    add_source_to_id=True,
                )

    def _set_tags(self):
        self._exposure_tag = self._exposure_tag_combo.currentText()
        self._x_tag = self._x_tag_combo.currentText()
        self._y_tag = self._y_tag_combo.currentText()
        self._column = self._column_combo.currentText()
        self._set_id_tag = self._set_id_combo.currentText()

    def load(self):
        self._set_tags()
        if (
            self._exposure_tag_combo.currentText()
            and self._x_tag_combo.currentText()
            and self._y_tag_combo.currentText()
            and self._column_combo.currentText()
            and self._set_id_combo.currentText()
            and self._group_name_box.text()
        ):

            if "*" in self._file_combo.currentText():
                thread_pool = ThreadPoolExecutor(max_workers=10)
                futures = []
                if "*" in self._cross_ref_file_combo.currentText():
                    split_file_name = self._file_combo.currentText().split("*")
                    for sfp in _string_to_glob(self._file_combo.currentText()):
                        if not sfp.parent.is_symlink():
                            wildcard = (
                                str(sfp)
                                .replace(split_file_name[0], "")
                                .replace(split_file_name[-1], "")
                            )
                            futures.append(
                                thread_pool.submit(
                                    self._insert_from_star_file,
                                    sfp,
                                    cross_ref_file_path=Path(
                                        self._cross_ref_file_combo.currentText().replace(
                                            "*", wildcard
                                        )
                                    ),
                                    new_thread=True,
                                )
                            )
                elif self._cross_ref_file_combo.currentText():
                    for sfp in _string_to_glob(self._file_combo.currentText()):
                        if not sfp.parent.is_symlink():
                            self._insert_from_star_file(
                                Path(sfp),
                                cross_ref_file_path=Path(
                                    self._cross_ref_file_combo.currentText()
                                ),
                            )
                else:
                    for sfp in _string_to_glob(self._file_combo.currentText()):
                        if not sfp.parent.is_symlink():
                            futures.append(
                                thread_pool.submit(
                                    self._insert_from_star_file, Path(sfp)
                                )
                            )
                for t in futures:
                    t.result()
            elif self._cross_ref_file_combo.currentText():
                self._insert_from_star_file(
                    Path(self._file_combo.currentText()),
                    cross_ref_file_path=Path(self._cross_ref_file_combo.currentText()),
                )
            else:
                self._insert_from_star_file(Path(self._file_combo.currentText()))
            self.refresh()



from pathlib import Path
from typing import List, Optional

from numpy import argmin

from smartem.data_model.extract import DataAPI
from smartem.parsing.star import (
    get_column_data,
    insert_exposure_data,
    insert_particle_data,
    insert_particle_set,
    open_star_file,
)


def _motion_corr(relion_dir: Path, data_handler: DataAPI, project: str):
    mc_base_path = relion_dir / "MotionCorr"
    mc_file_job_paths = mc_base_path.glob("job*")
    job_numbers = [str(j.name).replace("job", "") for j in mc_file_job_paths]
    first_job_idx = argmin([int(jn) for jn in job_numbers])
    mc_file_path = (
        relion_dir
        / "MotionCorr"
        / f"job{job_numbers[first_job_idx]}"
        / "corrected_micrographs.star"
    )
    star_file = open_star_file(mc_file_path)
    column_data = get_column_data(
        star_file, ["_rlnmicrographname", "_rlnaccummotiontotal"], "micrographs"
    )
    insert_exposure_data(
        column_data,
        "_rlnmicrographname",
        str(mc_file_path),
        data_handler,
        project=project,
    )


def _ctf(relion_dir: Path, data_handler: DataAPI, project: str):
    ctf_base_path = relion_dir / "CtfFind"
    ctf_file_job_paths = ctf_base_path.glob("job*")
    job_numbers = [str(j.name).replace("job", "") for j in ctf_file_job_paths]
    first_job_idx = argmin([int(jn) for jn in job_numbers])
    ctf_file_path = (
        relion_dir
        / "CtfFind"
        / f"job{job_numbers[first_job_idx]}"
        / "micrographs_ctf.star"
    )
    star_file = open_star_file(ctf_file_path)
    column_data = get_column_data(
        star_file, ["_rlnmicrographname", "_rlnctfmaxresolution"], "micrographs"
    )
    insert_exposure_data(
        column_data,
        "_rlnmicrographname",
        str(ctf_file_path),
        data_handler,
        project=project,
    )


def _prob_dist_max_class2d(
    relion_dir: Path,
    data_handler: DataAPI,
    project: str,
    excluded_directories: Optional[List[str]] = None,
):
    exclude = excluded_directories or []
    for class_file_path in (relion_dir / "Class2D").glob("job*"):
        if class_file_path.name not in exclude:
            star_file = open_star_file(class_file_path / "run_it020_data.star")
            column_data = get_column_data(
                star_file,
                [
                    "_rlnmicrographname",
                    "_rlncoordinatex",
                    "_rlncoordinatey",
                    "_rlnmaxvalueprobdistribution",
                ],
                "particles",
            )
            insert_particle_data(
                column_data,
                "_rlnmicrographname",
                "_rlncoordinatex",
                "_rlncoordinatey",
                str(class_file_path / "run_it020_data.star"),
                data_handler,
                project=project,
            )


def _class2d(
    relion_dir: Path,
    data_handler: DataAPI,
    project: str,
    excluded_directories: Optional[List[str]] = None,
):
    exclude = excluded_directories or []
    for class_file_path in (relion_dir / "Class2D").glob("job*"):
        if class_file_path.name not in exclude:
            star_file = open_star_file(class_file_path / "run_it020_data.star")
            cross_ref_star_file = open_star_file(
                class_file_path / "run_it020_model.star"
            )
            column_data = get_column_data(
                star_file,
                [
                    "_rlnmicrographname",
                    "_rlncoordinatex",
                    "_rlncoordinatey",
                    "_rlnclassnumber",
                ],
                "particles",
            )
            cross_ref_column_data = get_column_data(
                cross_ref_star_file,
                ["_rlnreferenceimage", "_rlnestimatedresolution"],
                "model_classes",
            )
            for v in cross_ref_column_data.values():
                for i, elem in enumerate(v):
                    if isinstance(elem, str) and "@" in elem:
                        _elem = elem.split("@")[0]
                        while _elem.startswith("0"):
                            _elem = _elem[1:]
                        v[i] = _elem
            cross_ref_dict = {}
            for i, k02 in enumerate(column_data["_rlnclassnumber"]):
                for j, k01 in enumerate(cross_ref_column_data["_rlnreferenceimage"]):
                    if k01 == str(k02):
                        cross_ref_dict[i] = cross_ref_column_data[
                            "_rlnestimatedresolution"
                        ][j]
            column_data["_rlnestimatedresolution"] = [
                cross_ref_dict[i] for i in range(len(column_data["_rlnclassnumber"]))
            ]
            insert_particle_set(
                column_data,
                "class_2d",
                "_rlnclassnumber",
                "_rlnmicrographname",
                "_rlncoordinatex",
                "_rlncoordinatey",
                str(class_file_path / "run_it020_data.star"),
                data_handler,
                project,
                add_source_to_id=True,
            )


def gather_relion_defaults(
    relion_dir: Path,
    data_handler: DataAPI,
    project: str,
    class_2d_excludes: Optional[List[str]] = None,
):
    _motion_corr(relion_dir, data_handler, project)
    _ctf(relion_dir, data_handler, project)
    _prob_dist_max_class2d(
        relion_dir, data_handler, project, excluded_directories=class_2d_excludes
    )
    _class2d(relion_dir, data_handler, project, excluded_directories=class_2d_excludes)



import csv
from pathlib import Path
from typing import Dict, List, Optional

from PyQt5.QtWidgets import (
    QComboBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
)

from smartem.data_model.extract import DataAPI
from smartem.gui.qt.component_tab import ComponentTab, background
from smartem.parsing.star import insert_exposure_data


class CSVDataLoader(ComponentTab):
    def __init__(
        self,
        extractor: DataAPI,
        project_directory: Optional[Path] = None,
        refreshers: Optional[List[ComponentTab]] = None,
    ):
        super().__init__(refreshers=refreshers)
        self._extractor = extractor
        self.grid = QGridLayout()
        self.setLayout(self.grid)
        self._exposure_tag = None
        self._column = None
        self._proj_dir = project_directory
        self.project = ""

        csv_lbl = QLabel()
        csv_lbl.setText("CSV file:")
        column_lbl = QLabel()
        column_lbl.setText("Data column:")

        self._file_combo = QComboBox()
        self._file_combo.setEditable(True)
        self._file_combo.currentIndexChanged.connect(self._select_csv_file)

        csv_hbox = QHBoxLayout()
        csv_hbox.addWidget(csv_lbl, 1)
        csv_hbox.addWidget(self._file_combo, 1)

        self._file_vbox = QVBoxLayout()
        self._file_vbox.addLayout(csv_hbox, 1)

        self.grid.addLayout(self._file_vbox, 2, 1, 1, 2)
        self._column_combo = QComboBox()
        self._column_combo.setEditable(True)
        self._column_combo.currentIndexChanged.connect(self._select_column)

        column_hbox = QHBoxLayout()
        column_hbox.addWidget(column_lbl, 1)
        column_hbox.addWidget(self._column_combo, 1)

        self.grid.addLayout(column_hbox, 3, 1, 1, 2)

        load_btn = QPushButton("Load")
        load_btn.clicked.connect(self.load)
        self.grid.addWidget(load_btn, 5, 1)

        exposure_lbl = QLabel()
        exposure_lbl.setText("Micrograph identifier:")

        self._exposure_tag_combo = QComboBox()
        self._exposure_tag_combo.setEditable(True)
        self._exposure_tag_combo.currentIndexChanged.connect(self._select_exposure_tag)

        self._identifier_box = QVBoxLayout()

        hbox = QHBoxLayout()
        hbox.addWidget(exposure_lbl, 1)
        hbox.addWidget(self._exposure_tag_combo, 1)

        self._identifier_box.addLayout(hbox, 1)

        self.grid.addLayout(self._identifier_box, 4, 1, 1, 2)

        self._exposure_tag = self._exposure_tag_combo.currentText()

    def _select_exposure_tag(self, index: int):
        self._exposure_tag = self._exposure_tag_combo.currentText()

    @background(children=None)
    def _set_project_directory(self, project_directory: Path):
        self._proj_dir = project_directory
        for sf in self._proj_dir.glob("*/*/*.csv"):
            str_sf = str(sf)
            self._file_combo.addItem(str_sf)

    def _insert_from_csv_file(self, csv_file_path: Path):
        if self._exposure_tag and self._column:
            column_data: Dict[str, List[str]] = {
                self._exposure_tag: [],
                self._column: [],
            }
            with open(csv_file_path, newline="") as csv_file:
                reader = csv.DictReader(csv_file)
                for row in reader:
                    column_data[self._exposure_tag].append(row[self._exposure_tag])
                    column_data[self._column].append(row[self._column])
            insert_exposure_data(
                column_data,
                self._exposure_tag,
                str(csv_file_path),
                self._extractor,
                extra_suffix="_grouped",
                project=self.project,
            )

    def _select_csv_file(
        self,
        index: int,
        defaults: Optional[List[str]] = None,
        connections: Optional[Dict[str, str]] = None,
    ):
        csv_file_path = Path(self._file_combo.currentText())
        with open(csv_file_path, newline="") as csv_file:
            reader = csv.DictReader(csv_file)
            columns = list(next(reader).keys())
        column_combos = [self._column_combo, self._exposure_tag_combo]
        for combo in column_combos:
            combo.clear()
            combo.addItem("")
        for c in sorted(set(columns)):
            for i, combo in enumerate(column_combos):
                combo.addItem(c)
                if defaults and c == defaults[i]:
                    combo.setCurrentText(c)
                    if connections and connections.get(defaults[i]):
                        setattr(self, connections[defaults[i]], c)

    def _select_column(self, index: int):
        self._column = self._column_combo.currentText()

    def load(self, **kwargs):
        if (
            self._exposure_tag or self._exposure_tag_combo.currentText()
        ) and self._column:
            if not self._exposure_tag:
                self._exposure_tag = self._exposure_tag_combo.currentText()
            self._insert_from_csv_file(Path(self._file_combo.currentText()))
            self.refresh()



import functools
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple

import mrcfile
import numpy as np
import pandas as pd
import yaml
from PIL import Image
from torch import Tensor, reshape, zeros
from torch.utils.data import DataLoader
from torchvision.io import read_image

from smartem.data_model import EPUImage, FoilHole, GridSquare
from smartem.data_model.extract import DataAPI
from smartem.data_model.structure import (
    extract_keys_with_foil_hole_averages,
    extract_keys_with_grid_square_averages,
)
from smartem.stage_model import StageCalibration, find_point_pixel


@functools.lru_cache(maxsize=50)
def mrc_to_tensor(mrc_file: Path) -> Tensor:
    with mrcfile.open(mrc_file) as mrc:
        data = mrc.data
    shape = data.shape
    if data.dtype.char in np.typecodes["AllInteger"]:
        tensor_2d = Tensor(data.astype(np.int16))
    else:
        tensor_2d = Tensor(data.astype(np.float16))
    return reshape(tensor_2d, (1, shape[0], shape[1]))


class SmartEMDataLoader(DataLoader):
    def __init__(
        self,
        level: str,
        epu_dir: Path,
        project: str,
        atlas_id: int,
        data_api: DataAPI,
        mrc: bool = False,
    ):
        self._data_api = data_api
        self._level = level
        self._epu_dir = epu_dir
        self._mrc = mrc
        atlas_info = self._data_api.get_atlas_info(
            atlas_id,
            ["_rlnaccummotiontotal", "_rlnctfmaxresolution"],
            [],
            ["_rlnestimatedresolution"],
        )
        if self._level not in ("grid_square", "foil_hole"):
            raise ValueError(
                f"Unrecognised SmartEMDataLoader level {self._level}: accepted values are grid_sqaure or foil_hole"
            )
        self._indexed: Sequence[EPUImage] = []
        if self._level == "grid_square":
            _labels = extract_keys_with_grid_square_averages(
                atlas_info,
                ["_rlnaccummotiontotal", "_rlnctfmaxresolution"],
                [],
                ["_rlnestimatedresolution"],
            )
            self._labels = {k: v.averages for k, v in _labels.items() if v.averages}
            _gs_indexed: Sequence[GridSquare] = self._data_api.get_grid_squares(
                project=project
            )
            self._image_paths = {
                p.grid_square_name: p.thumbnail
                for p in _gs_indexed
                if p.grid_square_name
            }
            self._indexed = _gs_indexed
        elif self._level == "foil_hole":
            _labels = extract_keys_with_foil_hole_averages(
                atlas_info,
                ["_rlnaccummotiontotal", "_rlnctfmaxresolution"],
                [],
                ["_rlnestimatedresolution"],
            )
            self._labels = {k: v.averages for k, v in _labels.items() if v.averages}
            _fh_indexed: Sequence[FoilHole] = self._data_api.get_foil_holes()
            self._image_paths = {
                p.foil_hole_name: p.thumbnail for p in _fh_indexed if p.foil_hole_name
            }
            self._indexed = _fh_indexed

    def __len__(self) -> int:
        return len(self._image_paths)

    def __getitem__(self, idx: int) -> Tuple[Tensor, List[float]]:
        ordered_labels = [
            "_rlnaccummotiontotal",
            "_rlnctfmaxresolution",
            "_rlnestimatedresolution",
        ]
        if self._level == "grid_square":
            index_name = self._indexed[idx].grid_square_name  # type: ignore
        elif self._label == "foil_hole":
            index_name = self._indexed[idx].foil_hole_name  # type: ignore
        img_path = self._image_paths[index_name]
        if img_path:
            if self._mrc:
                image = mrc_to_tensor((self._epu_dir / img_path).with_suffix(".mrc"))
            else:
                image = read_image(str(self._epu_dir / img_path))
            labels = [self._labels[l][index_name] for l in ordered_labels]
        else:
            image = zeros(1, 512, 512)
        return image, labels


_standard_labels = {
    "accummotiontotal": True,
    "ctfmaxresolution": True,
    "estimatedresolution": True,
    "maxvalueprobdistribution": False,
}


class SmartEMDiskDataLoader(DataLoader):
    def __init__(
        self,
        level: str,
        data_dir: Path,
        mrc: bool = False,
        labels_csv: str = "labels.csv",
        num_samples: int = 0,
        sub_sample_size: Optional[Tuple[int, int]] = None,
        allowed_labels: Optional[Dict[str, bool]] = None,
        seed: int = 0,
    ):
        np.random.seed(seed)
        self._level = level
        self._data_dir = data_dir
        self._mrc = mrc
        self._num_samples = num_samples
        self._sub_sample_size = sub_sample_size or (256, 256)
        self._allowed_labels = allowed_labels or list(_standard_labels.keys())
        self._lower_better_label = (
            [allowed_labels[k] for k in self._allowed_labels]
            if allowed_labels
            else [_standard_labels[k] for k in self._allowed_labels]
        )
        if self._level not in ("grid_square", "foil_hole"):
            raise ValueError(
                f"Unrecognised SmartEMDataLoader level {self._level}: accepted values are grid_sqaure or foil_hole"
            )
        self._df = pd.read_csv(self._data_dir / labels_csv)
        with open(self._data_dir / "coordinate_calibration.yaml", "r") as cal_in:
            sc = yaml.safe_load(cal_in)
        self._stage_calibration = StageCalibration(**sc)
        if level == "foil_hole":
            self._df = self._df[self._df["foil_hole"].notna()]
        with mrcfile.open(
            (self._data_dir / self._df.iloc[0]["grid_square"]).with_suffix(".mrc")
        ) as _mrc:
            self._gs_mrc_size = _mrc.data.shape
        with Image.open(self._data_dir / self._df.iloc[0]["grid_square"]) as im:
            self._gs_jpeg_size = im.size
        for row in self._df:
            try:
                with mrcfile.open(
                    (self._data_dir / self._df.iloc[0]["foil_hole"]).with_suffix(".mrc")
                ) as _mrc:
                    self._fh_mrc_size = _mrc.data.shape
                with Image.open(self._data_dir / self._df.iloc[0]["foil_hole"]) as im:
                    self._fh_jpeg_size = im.size
                break
            except TypeError:
                continue
        if self._mrc:
            self._boundary_points_x = np.random.randint(
                self._gs_mrc_size[1] - self._sub_sample_size[0], size=len(self)
            )
            self._boundary_points_y = np.random.randint(
                self._gs_mrc_size[0] - self._sub_sample_size[1], size=len(self)
            )
        else:
            self._boundary_points_x = np.random.randint(
                self._gs_jpeg_size[0] - self._sub_sample_size[0], size=len(self)
            )
            self._boundary_points_y = np.random.randint(
                self._gs_jpeg_size[1] - self._sub_sample_size[1], size=len(self)
            )

    def __len__(self) -> int:
        if self._level == "grid_square" and self._num_samples:
            return self._df[self._level].nunique() * self._num_samples
        return self._df[self._level].nunique()

    def __getitem__(self, idx: int) -> Tuple[Tensor, List[float]]:
        sub_sample_boundaries = (-1, -1)
        if self._level == "grid_square" and self._num_samples:
            sub_sample_boundaries = (
                self._boundary_points_x[idx],
                self._boundary_points_y[idx],
            )
            grid_square_idx = idx // self._num_samples
            _grid_squares = self._df["grid_square"].unique()
            selected_df = self._df[
                self._df["grid_square"] == _grid_squares[grid_square_idx]
            ]
            drop_indices = []
            if self._mrc:
                for ri, row in selected_df.iterrows():
                    fh_centre = find_point_pixel(
                        (
                            row["foil_hole_x"],
                            row["foil_hole_y"],
                        ),
                        (row["grid_square_x"], row["grid_square_y"]),
                        row["grid_square_pixel_size"],
                        (self._gs_mrc_size[1], self._gs_mrc_size[0]),
                        xfactor=-1 if self._stage_calibration.x_flip else 1,
                        yfactor=-1 if self._stage_calibration.y_flip else 1,
                    )
                    if self._stage_calibration.inverted:
                        fh_centre = (fh_centre[1], fh_centre[0])
                    if (
                        fh_centre[0] < sub_sample_boundaries[0]
                        or fh_centre[1] < sub_sample_boundaries[1]
                        or fh_centre[0]
                        > sub_sample_boundaries[0] + self._sub_sample_size[0]
                        or fh_centre[1]
                        > sub_sample_boundaries[1] + self._sub_sample_size[1]
                    ):
                        drop_indices.append(ri)
            else:
                for ri, row in selected_df.iterrows():
                    fh_centre = find_point_pixel(
                        (
                            row["foil_hole_x"],
                            row["foil_hole_y"],
                        ),
                        (row["grid_square_x"], row["grid_square_y"]),
                        row["grid_square_pixel_size"]
                        * (self._gs_mrc_size[1] / self._gs_jpeg_size[0]),
                        self._gs_jpeg_size,
                        xfactor=-1 if self._stage_calibration.x_flip else 1,
                        yfactor=-1 if self._stage_calibration.y_flip else 1,
                    )
                    if self._stage_calibration.inverted:
                        fh_centre = (fh_centre[1], fh_centre[0])
                    if (
                        fh_centre[0] < sub_sample_boundaries[0]
                        or fh_centre[1] < sub_sample_boundaries[1]
                        or fh_centre[0]
                        > sub_sample_boundaries[0] + self._sub_sample_size[0]
                        or fh_centre[1]
                        > sub_sample_boundaries[1] + self._sub_sample_size[1]
                    ):
                        drop_indices.append(ri)
            selected_df = selected_df.drop(drop_indices)
            averaged_df = selected_df.groupby("grid_square").mean()
            if len(averaged_df):
                labels = [
                    v
                    for k, v in averaged_df.iloc[0].to_dict().items()
                    if k in self._allowed_labels
                ]
            else:
                labels = [np.inf if b else -np.inf for b in self._lower_better_label]
            if self._mrc:
                image = mrc_to_tensor(
                    (self._data_dir / _grid_squares[grid_square_idx]).with_suffix(
                        ".mrc"
                    )
                )[
                    :,
                    sub_sample_boundaries[1] : sub_sample_boundaries[1]
                    + self._sub_sample_size[1],
                    sub_sample_boundaries[0] : sub_sample_boundaries[0]
                    + self._sub_sample_size[0],
                ]
            else:
                image = read_image(
                    str(self._data_dir / _grid_squares[grid_square_idx])
                )[
                    :,
                    sub_sample_boundaries[1] : sub_sample_boundaries[1]
                    + self._sub_sample_size[1],
                    sub_sample_boundaries[0] : sub_sample_boundaries[0]
                    + self._sub_sample_size[0],
                ]
        elif self._level == "grid_square":
            averaged_df = self._df.groupby("grid_square").mean()
            labels = [
                v
                for k, v in averaged_df.iloc[idx].to_dict().items()
                if k in self._allowed_labels
            ]
            if self._mrc:
                image = mrc_to_tensor(
                    (self._data_dir / averaged_df.iloc[idx].name).with_suffix(".mrc")
                )
            else:
                image = read_image(str(self._data_dir / averaged_df.iloc[idx].name))
        else:
            labels = [
                v
                for k, v in self._df.iloc[idx].to_dict().items()
                if k in self._allowed_labels
            ]
            if self._mrc:
                image = mrc_to_tensor(
                    (self._data_dir / self._df.iloc[idx][self._level]).with_suffix(
                        ".mrc"
                    )
                )
            else:
                image = read_image(
                    str(self._data_dir / self._df.iloc[idx][self._level])
                )
        return image, labels



import os
from pathlib import Path
from typing import Any, List, Optional, Type, Union, cast

import yaml
from sqlalchemy import Column
from sqlalchemy import Float as Float_org
from sqlalchemy import ForeignKey, Integer, Text, create_engine, inspect
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.sql.type_api import TypeEngine

# this is a mypy workaround suggeted in https://github.com/dropbox/sqlalchemy-stubs/issues/178
Float = cast(Type[TypeEngine[float]], Float_org)

Base: Any = declarative_base()


class EPUImage:
    stage_position_x: Column = Column(
        Float,
        comment="x postion of the microscope stage [nm]",
        nullable=False,
    )

    stage_position_y: Column = Column(
        Float,
        comment="y postion of the microscope stage [nm]",
        nullable=False,
    )

    thumbnail = Column(
        Text,
        nullable=True,
        comment="Full path to EPU jpeg image of EPU image",
    )

    pixel_size: Column = Column(
        Float,
        nullable=True,
        comment="Pixel size of full readout image extracted from EPU XML [nm]",
    )

    readout_area_x = Column(
        Integer,
        nullable=True,
        comment="x-extent of detector readout area",
    )

    readout_area_y = Column(
        Integer,
        nullable=True,
        comment="y-extent of detector readout area",
    )


class Project(Base):
    __tablename__ = "Project"

    project_name = Column(
        Text,
        primary_key=True,
        nullable=False,
    )

    acquisition_directory = Column(
        Text,
        nullable=False,
    )

    acquisition_software = Column(
        Text,
        nullable=False,
    )

    acquisition_software_version = Column(
        Text,
        nullable=False,
    )

    processing_directory = Column(Text)

    atlas_id: Column = Column(ForeignKey("Atlas.atlas_id"), index=True)
    Atlas = relationship("Atlas")


class Atlas(EPUImage, Base):
    __tablename__ = "Atlas"

    atlas_id = Column(
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )


class Tile(EPUImage, Base):
    __tablename__ = "Tile"

    tile_id = Column(
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )

    atlas_id: Column = Column(ForeignKey("Atlas.atlas_id"), index=True)
    Atlas = relationship("Atlas")


class GridSquare(EPUImage, Base):
    __tablename__ = "GridSquare"

    grid_square_name = Column(
        Text,
        primary_key=True,
        nullable=False,
    )

    tile_id: Column = Column(ForeignKey("Tile.tile_id"), index=True)
    Tile = relationship("Tile")


class FoilHole(EPUImage, Base):
    __tablename__ = "FoilHole"

    foil_hole_name = Column(
        Text,
        primary_key=True,
        nullable=False,
    )

    adjusted_stage_position_x: Column = Column(
        Float,
        comment="x postion of the microscope stage adjusted to account for beam shift [nm]",
        nullable=True,
    )
    adjusted_stage_position_y: Column = Column(
        Float,
        comment="y postion of the microscope stage adjusted to account for beam shift [nm]",
        nullable=True,
    )

    grid_square_name: Column = Column(
        ForeignKey("GridSquare.grid_square_name"), index=True
    )
    GridSquare = relationship("GridSquare")


class Exposure(EPUImage, Base):
    __tablename__ = "Exposure"

    exposure_name = Column(
        Text,
        primary_key=True,
        nullable=False,
    )

    foil_hole_name: Column = Column(ForeignKey("FoilHole.foil_hole_name"), index=True)
    FoilHole = relationship("FoilHole")


class Particle(Base):
    __tablename__ = "Particle"

    particle_id = Column(
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )

    x: Column = Column(
        Float,
        nullable=False,
    )

    y: Column = Column(
        Float,
        nullable=False,
    )

    exposure_name: Column = Column(ForeignKey("Exposure.exposure_name"), index=True)
    Exposure = relationship("Exposure")


class InfoStore:
    source = Column(
        Text,
        primary_key=True,
        nullable=False,
    )

    key = Column(
        Text,
        primary_key=True,
        nullable=False,
    )

    value: Column = Column(
        Float,
        primary_key=True,
        nullable=False,
    )


class ParticleInfo(InfoStore, Base):
    __tablename__ = "ParticleInfo"

    particle_id: Column = Column(
        ForeignKey("Particle.particle_id"), primary_key=True, index=True
    )
    Particle = relationship("Particle")


class ExposureInfo(InfoStore, Base):
    __tablename__ = "ExposureInfo"

    exposure_name: Column = Column(
        ForeignKey("Exposure.exposure_name"), primary_key=True, index=True
    )
    Exposure = relationship("Exposure")


class ParticleSet(Base):
    __tablename__ = "ParticleSet"

    group_name = Column(
        Text,
        primary_key=True,
        nullable=False,
    )

    identifier = Column(
        Text,
        primary_key=True,
        nullable=False,
        unique=True,
    )

    project_name: Column = Column(
        ForeignKey("Project.project_name"), nullable=False, index=True
    )
    Project = relationship("Project")


class ParticleSetInfo(InfoStore, Base):
    __tablename__ = "ParticleSetInfo"

    set_name: Column = Column(
        ForeignKey("ParticleSet.identifier"), primary_key=True, index=True
    )
    ParticleSet = relationship("ParticleSet")


class ParticleSetLinker(Base):
    __tablename__ = "ParticleSetLinker"

    particle_id: Column = Column(
        ForeignKey("Particle.particle_id"), primary_key=True, index=True
    )
    Particle = relationship("Particle")

    set_name: Column = Column(
        ForeignKey("ParticleSet.identifier"), primary_key=True, index=True
    )
    ParticleSet = relationship("ParticleSet")


_tables: List[Type[Base]] = [
    Atlas,
    Project,
    Tile,
    GridSquare,
    FoilHole,
    Exposure,
    Particle,
    ParticleInfo,
    ParticleSet,
    ParticleSetInfo,
    ParticleSetLinker,
    ExposureInfo,
]


def url(credentials_file: Optional[Union[str, Path]] = None) -> str:
    if not credentials_file:
        credentials_file = os.getenv("SMARTEM_CREDENTIALS")

    if not credentials_file:
        raise AttributeError(
            "No credentials file specified for smartem database (environment variable SMARTEM_CREDENTIALS)"
        )

    with open(credentials_file, "r") as stream:
        creds = yaml.safe_load(stream)

    return f"postgresql+psycopg2://{creds['username']}:{creds['password']}@{creds['host']}:{creds['port']}/{creds['database']}"


def setup():
    engine = create_engine(url())
    for tab in _tables:
        if not inspect(engine).has_table(tab.__tablename__):
            tab.__table__.create(engine)


def teardown():
    engine = create_engine(url())
    for tab in _tables[::-1]:
        try:
            tab.__table__.drop(engine)
        except ProgrammingError:
            pass



from pathlib import Path
from typing import Any, Dict, Optional, Tuple

import xmltodict

from smartem.data_model import Atlas, Exposure, FoilHole, GridSquare, Tile
from smartem.data_model.extract import DataAPI
from smartem.stage_model import StageCalibration, calibrate


def parse_epu_xml(xml_path: Path) -> Dict[str, Any]:
    with open(xml_path, "r") as xml:
        for_parsing = xml.read()
        data = xmltodict.parse(for_parsing)
    data = data["MicroscopeImage"]
    stage_position = data["microscopeData"]["stage"]["Position"]
    readout_area = data["microscopeData"]["acquisition"]["camera"]["ReadoutArea"]
    return {
        "stage_position": (
            float(stage_position["X"]) * 1e9,
            float(stage_position["Y"]) * 1e9,
        ),
        "pixel_size": float(data["SpatialScale"]["pixelSize"]["x"]["numericValue"])
        * 1e9,
        "readout_area": (int(readout_area["a:width"]), int(readout_area["a:height"])),
    }


def parse_epu_xml_version(xml_path: Path) -> Dict[str, Any]:
    with open(xml_path, "r") as xml:
        for_parsing = xml.read()
        data = xmltodict.parse(for_parsing)
    data = data["MicroscopeImage"]
    software = data["microscopeData"]["core"]["ApplicationSoftware"]
    version = data["microscopeData"]["core"]["ApplicationSoftwareVersion"]
    return {
        "software": software,
        "version": version,
    }


def metadata_foil_hole_positions(xml_path: Path) -> Dict[str, Tuple[int, int]]:
    with open(xml_path, "r") as xml:
        for_parsing = xml.read()
        data = xmltodict.parse(for_parsing)
    data = data["GridSquareXml"]
    serialization_array = data["TargetLocations"]["TargetLocationsEfficient"][
        "a:m_serializationArray"
    ]
    required_key = ""
    for key in serialization_array.keys():
        if key.startswith("b:KeyValuePairOfintTargetLocation"):
            required_key = key
            break
    if not required_key:
        return {}
    fh_pix_positions = {}
    for fh_block in serialization_array[required_key]:
        pix_center = fh_block["b:value"]["PixelCenter"]
        fh_pix_positions[fh_block["b:key"]] = (
            int(float(pix_center["c:x"])),
            int(float(pix_center["c:y"])),
        )
    return fh_pix_positions


def metadata_foil_hole_stage(xml_path: Path) -> Dict[str, Tuple[float, float]]:
    with open(xml_path, "r") as xml:
        for_parsing = xml.read()
        data = xmltodict.parse(for_parsing)
    data = data["GridSquareXml"]
    serialization_array = data["TargetLocations"]["TargetLocationsEfficient"][
        "a:m_serializationArray"
    ]
    required_key = ""
    for key in serialization_array.keys():
        if key.startswith("b:KeyValuePairOfintTargetLocation"):
            required_key = key
            break
    if not required_key:
        return {}
    fh_stage_positions = {}
    for fh_block in serialization_array[required_key]:
        stage = fh_block["b:value"]["StagePosition"]
        fh_stage_positions[fh_block["b:key"]] = (
            float(stage["c:X"]) * 1e9,
            float(stage["c:Y"]) * 1e9,
        )
    return fh_stage_positions


def metadata_foil_hole_corrected_stage(
    xml_path: Path,
) -> Dict[str, tuple]:
    with open(xml_path, "r") as xml:
        for_parsing = xml.read()
        data = xmltodict.parse(for_parsing)
    data = data["GridSquareXml"]
    serialization_array = data["TargetLocations"]["TargetLocationsEfficient"][
        "a:m_serializationArray"
    ]
    required_key = ""
    for key in serialization_array.keys():
        if key.startswith("b:KeyValuePairOfintTargetLocation"):
            required_key = key
            break
    if not required_key:
        return {}
    fh_cs_positions: Dict[str, tuple] = {}
    for fh_block in serialization_array[required_key]:
        if fh_block["b:value"]["IsPositionCorrected"] == "true":
            corrected_stage = fh_block["b:value"]["CorrectedStagePosition"]
            fh_cs_positions[fh_block["b:key"]] = (
                float(corrected_stage["c:X"]) * 1e9,
                float(corrected_stage["c:Y"]) * 1e9,
            )
        else:
            fh_cs_positions[fh_block["b:key"]] = (None, None)
    return fh_cs_positions


def calibrate_coordinate_system(xml_path: Path) -> Optional[StageCalibration]:
    with open(xml_path, "r") as xml:
        for_parsing = xml.read()
        data = xmltodict.parse(for_parsing)
    data = data["GridSquareXml"]
    serialization_array = data["TargetLocations"]["TargetLocationsEfficient"][
        "a:m_serializationArray"
    ]
    required_key = ""
    for key in serialization_array.keys():
        if key.startswith("b:KeyValuePairOfintTargetLocation"):
            required_key = key
            break
    if not required_key:
        return None
    found = False
    for i in range(len(serialization_array[required_key]) - 1):
        pix_positions = (
            (
                int(
                    float(
                        serialization_array[required_key][i]["b:value"]["PixelCenter"][
                            "c:x"
                        ]
                    )
                ),
                int(
                    float(
                        serialization_array[required_key][i]["b:value"]["PixelCenter"][
                            "c:y"
                        ]
                    )
                ),
            ),
            (
                int(
                    float(
                        serialization_array[required_key][
                            (i + 10) % len(serialization_array[required_key])
                        ]["b:value"]["PixelCenter"]["c:x"]
                    )
                ),
                int(
                    float(
                        serialization_array[required_key][
                            (i + 10) % len(serialization_array[required_key])
                        ]["b:value"]["PixelCenter"]["c:y"]
                    )
                ),
            ),
        )
        if (
            serialization_array[required_key][i]["b:value"]["IsPositionCorrected"]
            == "true"
        ):
            phys_01 = (
                float(
                    serialization_array[required_key][i]["b:value"][
                        "CorrectedStagePosition"
                    ]["c:X"]
                ),
                float(
                    serialization_array[required_key][i]["b:value"][
                        "CorrectedStagePosition"
                    ]["c:Y"]
                ),
            )
        else:
            phys_01 = (
                float(
                    serialization_array[required_key][i]["b:value"]["StagePosition"][
                        "c:X"
                    ]
                ),
                float(
                    serialization_array[required_key][i]["b:value"]["StagePosition"][
                        "c:Y"
                    ]
                ),
            )
        if (
            serialization_array[required_key][
                (i + 10) % len(serialization_array[required_key])
            ]["b:value"]["IsPositionCorrected"]
            == "true"
        ):
            phys_02 = (
                float(
                    serialization_array[required_key][
                        (i + 10) % len(serialization_array[required_key])
                    ]["b:value"]["CorrectedStagePosition"]["c:X"]
                ),
                float(
                    serialization_array[required_key][
                        (i + 10) % len(serialization_array[required_key])
                    ]["b:value"]["CorrectedStagePosition"]["c:Y"]
                ),
            )
        else:
            phys_02 = (
                float(
                    serialization_array[required_key][
                        (i + 10) % len(serialization_array[required_key])
                    ]["b:value"]["StagePosition"]["c:X"]
                ),
                float(
                    serialization_array[required_key][
                        (i + 10) % len(serialization_array[required_key])
                    ]["b:value"]["StagePosition"]["c:Y"]
                ),
            )
        physical_positions = (phys_01, phys_02)
        pix_diff = (
            pix_positions[1][0] - pix_positions[0][0],
            pix_positions[1][1] - pix_positions[0][1],
        )
        physical_diff = (
            physical_positions[1][0] - physical_positions[0][0],
            physical_positions[1][1] - physical_positions[0][1],
        )
        if (
            all(physical_diff)
            and all(pix_diff)
            and pix_diff[0] > 100
            and pix_diff[1] > 100
        ):
            found = True
            break
    if not found:
        return None
    return calibrate(pix_positions, physical_positions)


def create_atlas_and_tiles(atlas_image: Path, extractor: DataAPI) -> int:
    atlas_data = parse_epu_xml(atlas_image.with_suffix(".xml"))
    atlas = [
        Atlas(
            stage_position_x=atlas_data["stage_position"][0],
            stage_position_y=atlas_data["stage_position"][1],
            thumbnail=str(atlas_image),
            pixel_size=atlas_data["pixel_size"],
            readout_area_x=atlas_data["readout_area"][0],
            readout_area_y=atlas_data["readout_area"][1],
        )
    ]
    pid = extractor.put(atlas)
    atlas_id = pid[0].atlas_id  # atlas[0].atlas_id
    if atlas_id is None:
        raise RuntimeError(f"Atlas record was not correctly inserted: {atlas_image}")
    tiles = []
    for tile in atlas_image.parent.glob("Tile_*.jpg"):
        tile_data = parse_epu_xml(tile.with_suffix(".xml"))
        tiles.append(
            Tile(
                atlas_id=atlas_id,
                stage_position_x=tile_data["stage_position"][0],
                stage_position_y=tile_data["stage_position"][1],
                thumbnail=str(tile),
                pixel_size=tile_data["pixel_size"],
                readout_area_x=tile_data["readout_area"][0],
                readout_area_y=tile_data["readout_area"][1],
            )
        )
    extractor.put(tiles)
    return atlas_id


def parse_epu_version(epu_path: Path) -> Tuple[str, str]:
    xml_glob = iter(epu_path.glob("GridSquare*/*.xml"))
    res = parse_epu_xml_version(next(xml_glob))
    return (res["software"], res["version"])


def parse_epu_dir(epu_path: Path, extractor: DataAPI, project: str):
    exposures = {}
    for grid_square_dir in epu_path.glob("GridSquare*"):
        if grid_square_dir.is_dir():
            foil_holes: Dict[str, FoilHole] = {}
            afis_foil_holes: Dict[str, FoilHole] = {}
            grid_square_jpeg = next(grid_square_dir.glob("*.jpg"))
            grid_square_data = parse_epu_xml(grid_square_jpeg.with_suffix(".xml"))
            metadata_path = epu_path.parent / "Metadata"
            foil_hole_stage = metadata_foil_hole_stage(
                metadata_path / f"{grid_square_dir.name}.dm"
            )
            tile_id = extractor.get_tile_id(grid_square_data["stage_position"], project)
            if tile_id is not None:
                extractor.put(
                    [
                        GridSquare(
                            grid_square_name=grid_square_dir.name,
                            stage_position_x=grid_square_data["stage_position"][0],
                            stage_position_y=grid_square_data["stage_position"][1],
                            thumbnail=str(grid_square_jpeg.relative_to(epu_path)),
                            pixel_size=grid_square_data["pixel_size"],
                            readout_area_x=grid_square_data["readout_area"][0],
                            readout_area_y=grid_square_data["readout_area"][1],
                            tile_id=tile_id,
                        )
                    ]
                )
            for foil_hole_jpeg in (grid_square_dir / "FoilHoles").glob("FoilHole*.jpg"):
                foil_hole_name = "_".join(foil_hole_jpeg.stem.split("_")[:2])
                if foil_holes.get(foil_hole_name):
                    thumbnail = foil_holes[foil_hole_name].thumbnail
                    if thumbnail:
                        if (
                            epu_path / thumbnail
                        ).stat().st_mtime > foil_hole_jpeg.stat().st_mtime:
                            continue
                foil_hole_data = parse_epu_xml(foil_hole_jpeg.with_suffix(".xml"))
                foil_hole_label = foil_hole_name.split("_")[1]
                foil_holes[foil_hole_name] = FoilHole(
                    grid_square_name=grid_square_dir.name,
                    stage_position_x=foil_hole_data["stage_position"][0],
                    stage_position_y=foil_hole_data["stage_position"][1],
                    thumbnail=str(foil_hole_jpeg.relative_to(epu_path)),
                    pixel_size=foil_hole_data["pixel_size"],
                    readout_area_x=foil_hole_data["readout_area"][0],
                    readout_area_y=foil_hole_data["readout_area"][0],
                    foil_hole_name=foil_hole_name,
                    adjusted_stage_position_x=foil_hole_stage[foil_hole_label][0],
                    adjusted_stage_position_y=foil_hole_stage[foil_hole_label][1],
                )
            extractor.put(list(foil_holes.values()))
            for exposure_jpeg in (grid_square_dir / "Data").glob("*.jpg"):
                exposure_data = parse_epu_xml(exposure_jpeg.with_suffix(".xml"))
                for fh_name in foil_holes.keys():
                    if fh_name in exposure_jpeg.name:
                        foil_hole_name = fh_name
                        break
                else:
                    foil_hole_name = exposure_jpeg.name.split("_Data")[0]
                    foil_hole_label = foil_hole_name.split("_")[1]
                    adjusted_stage_position = foil_hole_stage[foil_hole_label]
                    afis_foil_holes[foil_hole_name] = FoilHole(
                        grid_square_name=grid_square_dir.name,
                        stage_position_x=exposure_data["stage_position"][0],
                        stage_position_y=exposure_data["stage_position"][1],
                        thumbnail=None,
                        pixel_size=None,
                        readout_area_x=None,
                        readout_area_y=None,
                        foil_hole_name=foil_hole_name,
                        adjusted_stage_position_x=adjusted_stage_position[0],
                        adjusted_stage_position_y=adjusted_stage_position[1],
                    )
                exposures[exposure_jpeg] = Exposure(
                    exposure_name=exposure_jpeg.name,
                    foil_hole_name=foil_hole_name,
                    stage_position_x=exposure_data["stage_position"][0],
                    stage_position_y=exposure_data["stage_position"][1],
                    thumbnail=str(exposure_jpeg.relative_to(epu_path)),
                    pixel_size=exposure_data["pixel_size"],
                    readout_area_x=exposure_data["readout_area"][0],
                    readout_area_y=exposure_data["readout_area"][1],
                )
            extractor.put(list(afis_foil_holes.values()))
            extractor.put(list(exposures.values()))



import argparse
from pathlib import Path

from smartem.data_model.extract import DataAPI
from smartem.parsing.star import get_column_data, open_star_file


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-p",
        "--project",
        help="Project name",
        dest="project",
    )
    parser.add_argument(
        "-i",
        "--imported",
        help="Star file from Relion import job",
        dest="imported",
    )
    parser.add_argument(
        "-o",
        "--output",
        help="File to dump results to",
        dest="output",
        default="./missing_movies",
    )
    args = parser.parse_args()

    star_file = open_star_file(Path(args.imported))
    movies = get_column_data(star_file, ["_rlnmicrographmoviename"], "movies")

    data_api = DataAPI()
    data_api.set_project(args.project)

    exposures = data_api.get_exposures()

    print(
        f"{len(exposures)} movies found in EPU directory, {len(movies['_rlnmicrographmoviename'])} movies imported in Relion (percentage {100*len(exposures)/len(movies['_rlnmicrographmoviename'])})"
    )

    foil_holes = data_api.get_foil_holes()
    imported_foil_holes = {
        f"FoilHole_{Path(m).name.split('_')[1]}": m.split("/")[1]
        for m in movies["_rlnmicrographmoviename"]
    }

    print(
        f"{len(foil_holes)} foil holes found in EPU directory, {len(imported_foil_holes)} foil holes imported in Relion (percentage {100*len(foil_holes)/len(imported_foil_holes)})"
    )

    grid_squares = data_api.get_grid_squares()
    imported_grid_squares = {m.split("/")[1] for m in movies["_rlnmicrographmoviename"]}

    print(
        f"{len(grid_squares)} grid squares found in EPU directory, {len(imported_grid_squares)} grid squares imported in Relion (percentage {100*len(grid_squares)/len(imported_grid_squares)})"
    )

    diff = set(imported_foil_holes.keys()) - {fh.foil_hole_name for fh in foil_holes}
    with open(f"{args.output}_{args.project}.txt", "w") as ofile:
        for d in diff:
            ofile.write(f"{imported_foil_holes[d]}/FoilHoles/{d}*\n")



from __future__ import annotations

import matplotlib
import matplotlib.ticker as mticker
import mrcfile

matplotlib.use("Qt5Agg")
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import numpy as np
import PyQt5.QtCore as QtCore
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
from PyQt5.QtGui import QPixmap, QTransform
from PyQt5.QtWidgets import (
    QCheckBox,
    QComboBox,
    QGridLayout,
    QLabel,
    QListWidget,
    QPushButton,
    QSlider,
    QVBoxLayout,
)

from smartem.data_model import Atlas, Exposure, FoilHole, GridSquare
from smartem.data_model.extract import DataAPI
from smartem.data_model.structure import (
    extract_keys,
    extract_keys_with_foil_hole_averages,
    extract_keys_with_grid_square_averages,
)
from smartem.gui.qt.component_tab import ComponentTab
from smartem.gui.qt.image_utils import ImageLabel, ParticleImageLabel
from smartem.gui.qt.plotting_utils import InteractivePlot
from smartem.parsing.epu import calibrate_coordinate_system
from smartem.stage_model import StageCalibration


class MainDisplay(ComponentTab):
    def __init__(
        self,
        extractor: DataAPI,
        atlas_view: Optional[AtlasDisplay] = None,
        refreshers: Optional[List[ComponentTab]] = None,
    ):
        super().__init__(refreshers=refreshers)
        self._extractor = extractor
        self._epu_dir: Optional[Path] = None
        self._data_size: Optional[Tuple[int, int]] = None
        self._data: Dict[str, List[float]] = {}
        self._foil_hole_averages: Dict[str, Dict[str, float]] = {}
        self._particle_data: Dict[str, List[float]] = {}
        self._exposure_keys: List[str] = []
        self._particle_keys: List[str] = []
        self._particle_set_keys: List[str] = []
        self.grid = QGridLayout()
        self.setLayout(self.grid)
        self._square_combo = QComboBox()
        self._square_combo.setEditable(True)
        self._square_combo.currentIndexChanged.connect(self._select_square)
        self._foil_hole_combo = QComboBox()
        self._foil_hole_combo.setEditable(True)
        self._foil_hole_combo.currentIndexChanged.connect(self._select_foil_hole)
        self._exposure_combo = QComboBox()
        self._exposure_combo.setEditable(True)
        self._exposure_combo.currentIndexChanged.connect(self._select_exposure)
        self._data_combo = QComboBox()
        self._data_list = QListWidget()
        self._data_list.setSelectionMode(QListWidget.MultiSelection)
        self._pick_list = QListWidget()
        self._pick_list.setSelectionMode(QListWidget.MultiSelection)
        fh_fig = Figure()
        fh_fig.set_facecolor("gray")
        self._foil_hole_stats_fig = fh_fig.add_subplot(111)
        self._foil_hole_stats_fig.set_facecolor("silver")
        self._foil_hole_stats = FigureCanvasQTAgg(fh_fig)
        gs_fig = Figure()
        gs_fig.set_facecolor("gray")
        self._grid_square_stats_fig = gs_fig.add_subplot(111)
        self._grid_square_stats_fig.set_facecolor("silver")
        self._grid_square_stats = FigureCanvasQTAgg(gs_fig)
        ex_fig = Figure()
        ex_fig.set_facecolor("gray")
        self._exposure_stats_fig = ex_fig.add_subplot(111)
        self._exposure_stats_fig.set_facecolor("silver")
        self._exposure_stats = FigureCanvasQTAgg(ex_fig)
        self.grid.addWidget(self._square_combo, 2, 1)
        self.grid.addWidget(self._foil_hole_combo, 2, 2)
        self.grid.addWidget(self._exposure_combo, 2, 3)
        self.grid.addWidget(self._data_list, 3, 2)
        self.grid.addWidget(self._pick_list, 3, 3)
        self.grid.addWidget(self._grid_square_stats, 4, 1)
        self.grid.addWidget(self._foil_hole_stats, 4, 2)
        self.grid.addWidget(self._exposure_stats, 4, 3)
        self._grid_squares: List[GridSquare] = []
        self._foil_holes: List[FoilHole] = []
        self._exposures: List[Exposure] = []
        self._atlas_view = atlas_view
        self._colour_bar = None
        self._fh_colour_bar = None
        self._exp_colour_bar = None
        self._data_gathered = False
        self._data_keys: Dict[str, List[str]] = {
            "micrograph": [],
            "particle": [],
            "particle_set": [],
        }
        self._pick_keys: Dict[str, List[str]] = {
            "source": [],
            "set_group": [],
        }

        self._gather_btn = QPushButton("Gather data")
        self._gather_btn.clicked.connect(self._gather_data)
        self.grid.addWidget(self._gather_btn, 3, 1)

        self._square_slider = QSlider(QtCore.Qt.Horizontal)
        self._square_slider.setMaximum(100)
        self._square_slider.setTickInterval(1)
        self._square_slider.setValue(100)
        self._square_slider_maximum = 0.0
        self._for_removal = None
        self._square_slider.valueChanged.connect(self._square_slider_changed)
        self.grid.addWidget(self._square_slider, 5, 1)

        self._particle_diameter_slider = QSlider(QtCore.Qt.Vertical)
        self._particle_diameter_slider.setMaximum(100)
        self._particle_diameter_slider.setTickInterval(1)
        self._particle_diameter_slider.setValue(30)
        self._particle_diameter = 30
        self._particle_diameter_slider.valueChanged.connect(
            self._particle_diameter_slider_changed
        )
        self.grid.addWidget(self._particle_diameter_slider, 1, 4)

        self._particle_tick_boxes = (
            QCheckBox("Show particles"),
            QCheckBox("Flip x"),
            QCheckBox("Flip y"),
        )
        self._particle_tick_boxes[0].toggled.connect(self._draw_current_exposure)
        self._particle_tick_boxes[1].toggled.connect(self._draw_current_exposure)
        self._particle_tick_boxes[2].toggled.connect(self._draw_current_exposure)
        self._particle_diameter_label = QLabel("Diameter: \n? Angstroms")
        # self._particle_dia
        particles_vbox = QVBoxLayout()
        full_vbox = QVBoxLayout()
        particles_vbox.addWidget(self._particle_tick_boxes[0], 1)
        particles_vbox.addWidget(self._particle_tick_boxes[1], 2)
        particles_vbox.addWidget(self._particle_tick_boxes[2], 3)
        full_vbox.addLayout(particles_vbox, 1)
        full_vbox.addWidget(self._particle_diameter_label, 2)
        self.grid.addLayout(full_vbox, 1, 5)

        self.project = ""
        self._stage_calibration = StageCalibration()

    def load(self):
        self._grid_squares = self._extractor.get_grid_squares(project=self.project)
        self._square_combo.clear()
        for gs in self._grid_squares:
            self._square_combo.addItem(gs.grid_square_name)
        self._update_fh_choices(self._grid_squares[0].grid_square_name)
        self.refresh()

    def set_project(self, project: str):
        self.project = project
        _project = self._extractor.get_project(project_name=self.project)
        _epu_version = _project.acquisition_software_version
        if (
            int(_epu_version.split(".")[0]) >= 2
            and int(_epu_version.split(".")[1]) > 12
        ):
            self._particle_tick_boxes[2].setChecked(True)

    def _set_epu_directory(self, epu_dir: Path):
        self._epu_dir = epu_dir
        for dm in (epu_dir.parent / "Metadata").glob("*.dm"):
            cal = calibrate_coordinate_system(dm)
            if cal:
                self._stage_calibration = cal
                break

    def _set_data_size(self, project_dir: Path):
        try:
            mcdir = project_dir / "MotionCorr" / "job002" / "Movies"
            first_mrc = next(iter(mcdir.glob("**/*.mrc")))
            with mrcfile.open(first_mrc) as mrc:
                self._data_size = mrc.data.shape
        except Exception:
            return

    def _gather_atlas_data(self):
        _grid_square = self._grid_squares[self._square_combo.currentIndex()]
        _atlas = self._extractor.get_atlases(project=self.project)
        atlas_sql_data = self._extractor.get_atlas_info(
            _atlas.atlas_id,
            self._exposure_keys,
            self._particle_keys,
            self._particle_set_keys,
        )
        extracted_atlas_data = extract_keys_with_grid_square_averages(
            atlas_sql_data,
            self._exposure_keys,
            self._particle_keys,
            self._particle_set_keys,
        )
        atlas_data = {k: v.flattened_data for k, v in extracted_atlas_data.items()}
        grid_square_averages = {k: v.averages for k, v in extracted_atlas_data.items()}
        if self._atlas_view and self._epu_dir:
            self._atlas_view._data = atlas_data
            self._atlas_view._grid_square_averages = grid_square_averages
            self._atlas_view.load(
                self._epu_dir,
                grid_square=_grid_square,
                all_grid_squares=self._grid_squares,
                data_changed=True,
            )

    def _gather_grid_square_data(self):
        foil_hole_exposures = {}
        for fh in self._foil_holes:
            foil_hole_exposures[fh.foil_hole_name] = [
                e.exposure_name
                for e in self._extractor.get_exposures(foil_hole_name=fh.foil_hole_name)
            ]

        sql_data = self._extractor.get_grid_square_info(
            self._square_combo.currentText(),
            self._exposure_keys,
            self._particle_keys,
            self._particle_set_keys,
        )
        extracted_grid_square_data = extract_keys_with_foil_hole_averages(
            sql_data,
            self._exposure_keys,
            self._particle_keys,
            self._particle_set_keys,
        )
        self._data = {
            k: v.flattened_data for k, v in extracted_grid_square_data.items()
        }
        self._foil_hole_averages = {
            k: v.averages for k, v in extracted_grid_square_data.items()
        }

    def _square_slider_changed(self, value: int):
        # print((value/100)*self._square_slider_maximum)
        try:
            self._draw_grid_square(
                self._grid_squares[self._square_combo.currentIndex()],
                foil_hole=self._foil_holes[self._foil_hole_combo.currentIndex()],
                max_value=(value / 100) * self._square_slider_maximum,
            )
        except IndexError:
            self._draw_grid_square(
                self._grid_squares[self._square_combo.currentIndex()],
                max_value=(value / 100) * self._square_slider_maximum,
            )
        if self._for_removal is not None:
            self._for_removal.remove()
        self._for_removal = self._grid_square_stats_fig.axvline(
            (value / 100) * self._square_slider_maximum
        )
        self._grid_square_stats.draw()

    def _particle_diameter_slider_changed(self, value: int):
        self._particle_diameter = value
        self._draw_current_exposure()
        exposure_pixmap = QPixmap(
            str(
                self._epu_dir
                / self._exposures[self._exposure_combo.currentIndex()].thumbnail
            )
        )
        qsize = exposure_pixmap.size()
        scaled_pixel_size = self._exposures[
            self._exposure_combo.currentIndex()
        ].pixel_size * (
            self._exposures[self._exposure_combo.currentIndex()].readout_area_x
            / qsize.width()
        )
        self._particle_diameter_label.setText(
            f"Diameter: \n{int(self._particle_diameter*scaled_pixel_size*10)} Angstroms"
        )

    def _gather_foil_hole_data(self):
        sql_data = self._extractor.get_foil_hole_info(
            self._foil_hole_combo.currentText(),
            self._exposure_keys,
            self._particle_keys,
            self._particle_set_keys,
        )
        key_extracted_data = extract_keys(
            sql_data,
            self._exposure_keys,
            self._particle_keys,
            self._particle_set_keys,
        )
        try:
            self._update_foil_hole_stats(key_extracted_data)
        except KeyError:
            pass

    def _gather_data(self, evt):
        self._square_slider.setValue(100)
        selected_keys = [d.text() for d in self._data_list.selectedItems()]
        self._exposure_keys = [
            k for k in selected_keys if k in self._data_keys["micrograph"]
        ]
        self._particle_keys = [
            k for k in selected_keys if k in self._data_keys["particle"]
        ]
        self._particle_set_keys = [
            k for k in selected_keys if k in self._data_keys["particle_set"]
        ]

        self._gather_grid_square_data()

        self._gather_foil_hole_data()

        self._data_gathered = True

        self._draw_foil_hole(
            self._foil_holes[self._foil_hole_combo.currentIndex()], flip=(-1, -1)
        )
        self._draw_exposure(
            self._exposures[self._exposure_combo.currentIndex()],
            flip=(
                -1 if self._particle_tick_boxes[1].isChecked() else 1,
                -1 if self._particle_tick_boxes[2].isChecked() else 1,
            ),
        )

        try:
            self._draw_grid_square(
                self._grid_squares[self._square_combo.currentIndex()],
                foil_hole=self._foil_holes[self._foil_hole_combo.currentIndex()],
            )
        except IndexError:
            self._draw_grid_square(
                self._grid_squares[self._square_combo.currentIndex()]
            )
        self._update_grid_square_stats(self._data)

        self._gather_atlas_data()

    def _select_square(self, index: int):
        if self._data_gathered:
            self._gather_grid_square_data()
            self._update_grid_square_stats(self._data)
        try:
            square_lbl = self._draw_grid_square(
                self._grid_squares[index],
                foil_hole=self._foil_holes[self._foil_hole_combo.currentIndex()],
            )
        except IndexError:
            try:
                square_lbl = self._draw_grid_square(self._grid_squares[index])
            except IndexError:
                return
        self.grid.addWidget(square_lbl, 1, 1)
        self._update_fh_choices(self._square_combo.currentText())

        if self._atlas_view and self._epu_dir:
            self._atlas_view.load(
                self._epu_dir,
                grid_square=self._grid_squares[index],
                all_grid_squares=self._grid_squares,
            )

    def _select_foil_hole(self, index: int):
        try:
            hole_lbl = self._draw_foil_hole(self._foil_holes[index], flip=(-1, -1))
        except IndexError:
            return
        self.grid.addWidget(hole_lbl, 1, 2)
        self._update_exposure_choices(self._foil_hole_combo.currentText())
        self._draw_grid_square(
            self._grid_squares[self._square_combo.currentIndex()],
            foil_hole=self._foil_holes[index],
        )
        if (
            any([self._exposure_keys, self._particle_keys, self._particle_set_keys])
            and self._foil_hole_combo.currentText()
        ):
            self._gather_foil_hole_data()

    def _update_grid_square_stats(self, stats: Dict[str, List[float]]):
        gs_fig = Figure(tight_layout=True)
        gs_fig.set_facecolor("gray")
        self._grid_square_stats_fig = gs_fig.add_subplot(111)
        self._grid_square_stats_fig.set_facecolor("silver")
        self._grid_square_stats = InteractivePlot(gs_fig)
        self.grid.addWidget(self._grid_square_stats, 4, 1)
        try:
            if self._colour_bar:
                self._colour_bar.remove()
        except (AttributeError, ValueError):
            pass
        if len(stats.keys()) == 1:
            self._grid_square_stats.set_data(list(stats.values())[0])
            self._grid_square_stats_fig.hist(
                list(stats.values())[0], color="darkturquoise"
            )
            self._grid_square_stats_fig.axes.set_xlabel(list(stats.keys())[0])
            self._square_slider_maximum = max(list(stats.values())[0])
        if len(stats.keys()) == 2:
            labels = []
            data = []
            for k, v in stats.items():
                labels.append(k)
                data.append(v)
            self._grid_square_stats.set_data(data)
            self._grid_square_stats_fig.scatter(
                data[0],
                data[1],
                color="darkturquoise",
            )
            self._grid_square_stats_fig.axes.set_xlabel(labels[0])
            self._grid_square_stats_fig.axes.set_ylabel(labels[1])
        if len(stats.keys()) > 2:
            labels = []
            data = []
            for k, v in stats.items():
                labels.append(k)
                data.append(list(v))
            for i, _ in enumerate(data):
                data[i] = np.nan_to_num(_)
            corr = np.corrcoef(data)
            mat = self._grid_square_stats_fig.matshow(corr)
            ticks_loc = (
                self._grid_square_stats_fig.axes.get_xticks(),
                self._grid_square_stats_fig.axes.get_yticks(),
            )
            self._grid_square_stats.set_data(corr)
            self._grid_square_stats_fig.xaxis.set_major_locator(
                mticker.FixedLocator(ticks_loc[0][1:-1])
            )
            self._grid_square_stats_fig.yaxis.set_major_locator(
                mticker.FixedLocator(ticks_loc[1][1:-1])
            )
            self._grid_square_stats_fig.axes.set_xticklabels(labels, rotation=45)
            self._grid_square_stats_fig.axes.set_yticklabels(labels)
            self._colour_bar = self._grid_square_stats_fig.figure.colorbar(mat)
        self._grid_square_stats.draw()

    def _update_foil_hole_stats(self, stats: Dict[str, List[float]]):
        fh_fig = Figure(tight_layout=True)
        fh_fig.set_facecolor("gray")
        self._foil_hole_stats_fig = fh_fig.add_subplot(111)
        self._foil_hole_stats_fig.set_facecolor("silver")
        self._foil_hole_stats = InteractivePlot(fh_fig)
        self.grid.addWidget(self._foil_hole_stats, 4, 2)
        try:
            if self._fh_colour_bar:
                self._fh_colour_bar.remove()
        except (AttributeError, ValueError):
            pass
        if len(stats.keys()) == 1:
            self._foil_hole_stats.set_data(list(stats.values())[0])
            self._foil_hole_stats_fig.hist(
                list(stats.values())[0], color="darkturquoise"
            )
            self._foil_hole_stats_fig.axes.set_xlabel(list(stats.keys())[0])
        if len(stats.keys()) == 2:
            labels = []
            data = []
            for k, v in stats.items():
                labels.append(k)
                data.append(v)
            self._foil_hole_stats.set_data(data)
            self._foil_hole_stats_fig.scatter(data[0], data[1], color="darkturquoise")
            self._foil_hole_stats_fig.axes.set_xlabel(labels[0])
            self._foil_hole_stats_fig.axes.set_ylabel(labels[1])
        if len(stats.keys()) > 2:
            labels = []
            data = []
            for k, v in stats.items():
                labels.append(k)
                data.append(list(v))
            corr = np.corrcoef(data)
            mat = self._foil_hole_stats_fig.matshow(corr)
            ticks_loc = (
                self._foil_hole_stats_fig.axes.get_xticks(),
                self._foil_hole_stats_fig.axes.get_yticks(),
            )
            self._foil_hole_stats.set_data(corr)
            self._foil_hole_stats_fig.xaxis.set_major_locator(
                mticker.FixedLocator(ticks_loc[0][1:-1])
            )
            self._foil_hole_stats_fig.yaxis.set_major_locator(
                mticker.FixedLocator(ticks_loc[1][1:-1])
            )
            self._foil_hole_stats_fig.axes.set_xticklabels(labels, rotation=45)
            self._foil_hole_stats_fig.axes.set_yticklabels(labels)
            self._fh_colour_bar = self._foil_hole_stats_fig.figure.colorbar(mat)
        self._foil_hole_stats.draw()

    def _update_foil_hole_stats_picks(self, stats: Dict[str, List[int]]):
        if len(stats.keys()) == 2:
            size_lists = list(stats.values())
            diffs = [p2 - p1 for p1, p2 in zip(size_lists[0], size_lists[1])]
            self._foil_hole_stats_fig.hist(diffs)
            self._foil_hole_stats.draw()

    def _update_exposure_stats(self, stats: Dict[str, List[float]]):
        ex_fig = Figure(tight_layout=True)
        ex_fig.set_facecolor("gray")
        self._exposure_stats_fig = ex_fig.add_subplot(111)
        self._exposure_stats_fig.set_facecolor("silver")
        self._exposure_stats = InteractivePlot(ex_fig)
        self.grid.addWidget(self._exposure_stats, 4, 3)
        try:
            if self._exp_colour_bar:
                self._exp_colour_bar.remove()
        except (AttributeError, ValueError):
            pass
        if len(stats.keys()) == 1:
            self._exposure_stats.set_data(list(stats.values())[0])
            self._exposure_stats_fig.hist(
                list(stats.values())[0], color="darkturquoise"
            )
            self._exposure_stats_fig.axes.set_xlabel(list(stats.keys())[0])
        if len(stats.keys()) == 2:
            labels = []
            data = []
            for k, v in stats.items():
                labels.append(k)
                data.append(v)
            self._exposure_stats.set_data(data)
            self._exposure_stats_fig.scatter(data[0], data[1], color="darkturquoise")
            self._exposure_stats_fig.axes.set_xlabel(labels[0])
            self._exposure_stats_fig.axes.set_ylabel(labels[1])
        if len(stats.keys()) > 2:
            labels = []
            data = []
            for k, v in stats.items():
                labels.append(k)
                data.append(list(v))
            corr = np.corrcoef(data)
            mat = self._exposure_stats_fig.matshow(corr)
            ticks_loc = (
                self._exposure_stats_fig.axes.get_xticks(),
                self._exposure_stats_fig.axes.get_yticks(),
            )
            self._exposure_stats.set_data(corr)
            self._exposure_stats_fig.xaxis.set_major_locator(
                mticker.FixedLocator(ticks_loc[0][1:-1])
            )
            self._exposure_stats_fig.yaxis.set_major_locator(
                mticker.FixedLocator(ticks_loc[1][1:-1])
            )
            self._exposure_stats_fig.axes.set_xticklabels(labels, rotation=45)
            self._exposure_stats_fig.axes.set_yticklabels(labels)
            self._exp_colour_bar = self._exposure_stats_fig.figure.colorbar(mat)

        self._exposure_stats.draw()

    def _draw_grid_square(
        self,
        grid_square: GridSquare,
        foil_hole: Optional[FoilHole] = None,
        flip: Tuple[int, int] = (1, 1),
        max_value: Optional[float] = None,
    ) -> QLabel:
        if not self._epu_dir or not grid_square.thumbnail:
            return
        square_pixmap = QPixmap(str(self._epu_dir / grid_square.thumbnail))
        if flip != (1, 1):
            square_pixmap = square_pixmap.transformed(QTransform().scale(*flip))
        if foil_hole and self._epu_dir:
            qsize = square_pixmap.size()
            imvs: Optional[list] = None
            _key = None
            if len(self._data.keys()) == 1:
                _key = list(self._data.keys())[0]
                imvs = [
                    list(self._foil_hole_averages.values())[0].get(fh.foil_hole_name)
                    for fh in self._foil_holes
                    if fh != foil_hole
                    and fh.thumbnail
                    or fh.adjusted_stage_position_x is not None
                ]
                if max_value is not None:
                    imvs = [i if i < max_value else None for i in imvs]
                    if (
                        list(self._foil_hole_averages.values())[0][
                            foil_hole.foil_hole_name
                        ]
                        > max_value
                    ):
                        _key = None
            square_lbl = ImageLabel(
                grid_square,
                foil_hole,
                (qsize.width(), qsize.height()),
                self._epu_dir,
                parent=self,
                value=list(self._foil_hole_averages.values())[0].get(
                    foil_hole.foil_hole_name
                )
                if _key
                else None,
                extra_images=[
                    fh
                    for fh in self._foil_holes
                    if fh != foil_hole
                    and fh.thumbnail
                    or fh.adjusted_stage_position_x is not None
                ],
                image_values=imvs,
                selection_box=self._square_combo,
                stage_calibration=self._stage_calibration,
            )
            self.grid.addWidget(square_lbl, 1, 1)
            square_lbl.setPixmap(square_pixmap)
        else:
            square_lbl = QLabel(self)
            square_lbl.setPixmap(square_pixmap)
        return square_lbl

    def _draw_foil_hole(
        self,
        foil_hole: FoilHole,
        exposure: Optional[Exposure] = None,
        flip: Tuple[int, int] = (1, 1),
    ) -> QLabel:
        if not self._epu_dir:
            return
        if foil_hole.thumbnail:
            hole_pixmap = QPixmap(str(self._epu_dir / foil_hole.thumbnail))
            if flip != (1, 1):
                hole_pixmap = hole_pixmap.transformed(QTransform().scale(*flip))
        if exposure and self._epu_dir:
            if foil_hole.thumbnail:
                qsize = hole_pixmap.size()
                hole_lbl = ImageLabel(
                    foil_hole,
                    exposure,
                    (qsize.width(), qsize.height()),
                    self._epu_dir,
                    parent=self,
                    selection_box=self._foil_hole_combo,
                    stage_calibration=self._stage_calibration,
                )
                self.grid.addWidget(hole_lbl, 1, 2)
                hole_lbl.setPixmap(hole_pixmap)
            else:
                hole_lbl = QLabel(self)
        else:
            hole_lbl = QLabel(self)
            if foil_hole.thumbnail:
                hole_lbl.setPixmap(hole_pixmap)
        return hole_lbl

    def _select_exposure(self, index: int):
        exposure_lbl = QLabel(self)
        try:
            exposure_lbl = self._draw_exposure(
                self._exposures[index],
                flip=(
                    -1 if self._particle_tick_boxes[1].isChecked() else 1,
                    -1 if self._particle_tick_boxes[2].isChecked() else 1,
                ),
            )
        except IndexError:
            return
        self.grid.addWidget(exposure_lbl, 1, 3)
        if self._foil_holes:
            self._draw_foil_hole(
                self._foil_holes[self._foil_hole_combo.currentIndex()],
                exposure=self._exposures[index],
                flip=(-1, -1),
            )
        if (
            any([self._particle_keys, self._particle_set_keys])
            and self._exposure_combo.currentText()
        ):

            sql_data = self._extractor.get_exposure_info(
                self._exposure_combo.currentText(),
                self._particle_keys,
                self._particle_set_keys,
            )
            key_extracted_data = extract_keys(
                sql_data,
                [],
                self._particle_keys,
                self._particle_set_keys,
            )
            self._update_exposure_stats(key_extracted_data)

    def _draw_current_exposure(self):
        try:
            self._draw_exposure(
                self._exposures[self._exposure_combo.currentIndex()],
                flip=(
                    -1 if self._particle_tick_boxes[1].isChecked() else 1,
                    -1 if self._particle_tick_boxes[2].isChecked() else 1,
                ),
            )
        except IndexError:
            return

    def _draw_exposure(
        self, exposure: Exposure, flip: Tuple[int, int] = (1, 1)
    ) -> QLabel:
        if not self._epu_dir or not exposure.thumbnail:
            return
        exposure_pixmap = QPixmap(str(self._epu_dir / exposure.thumbnail))
        if flip != (1, 1):
            exposure_pixmap = exposure_pixmap.transformed(QTransform().scale(*flip))
        qsize = exposure_pixmap.size()
        particles = []
        if self._particle_tick_boxes[0].isChecked():
            if self._pick_list.selectedItems():
                for p in self._pick_list.selectedItems():
                    if p.text() in self._pick_keys["source"]:
                        exp_parts = self._extractor.get_particles(
                            exposure_name=exposure.exposure_name, source=p.text()
                        )
                        particles.append(exp_parts)
                    else:
                        exp_parts = self._extractor.get_particles(
                            exposure_name=exposure.exposure_name,  # group_name=p.text()
                        )
                        particles.append(exp_parts)
            else:
                particles = [
                    self._extractor.get_particles(exposure_name=exposure.exposure_name)
                ]
        with mrcfile.open(
            (self._epu_dir / exposure.thumbnail).with_suffix(".mrc")
        ) as mrc:
            thumbnail_size = mrc.data.shape
        exposure_lbl = ParticleImageLabel(
            exposure,
            particles,
            (qsize.width(), qsize.height()),
            image_scale=0.5
            if self._data_size is None
            else thumbnail_size[0] / self._data_size[0],
            selection_box=self._exposure_combo,
            particle_diameter=self._particle_diameter,
        )
        self.grid.addWidget(exposure_lbl, 1, 3)
        exposure_lbl.setPixmap(exposure_pixmap)
        return exposure_lbl

    def _update_fh_choices(self, grid_square_name: str):
        self._foil_holes = self._extractor.get_foil_holes(
            grid_square_name=grid_square_name
        )
        self._foil_hole_combo.clear()
        for fh in self._foil_holes:
            self._foil_hole_combo.addItem(fh.foil_hole_name)

    def _update_exposure_choices(self, foil_hole_name: str):
        self._exposures = self._extractor.get_exposures(foil_hole_name=foil_hole_name)
        self._exposure_combo.clear()
        for ex in self._exposures:
            self._exposure_combo.addItem(ex.exposure_name)

    def refresh(self):
        super().refresh()
        self._data_list.clear()
        self._pick_list.clear()

        self._data_keys["micrograph"] = self._extractor.get_exposure_keys(self.project)
        self._data_keys["particle"] = self._extractor.get_particle_keys(self.project)
        self._data_keys["particle_set"] = self._extractor.get_particle_set_keys(
            self.project
        )
        for keys in self._data_keys.values():
            for k in keys:
                self._data_list.addItem(k)

        # self._pick_keys["source"] = self._extractor.get_particle_info_sources(
        #     self.project
        # )
        self._pick_keys["set_group"] = self._extractor.get_particle_set_group_names(
            self.project
        )
        for keys in self._pick_keys.values():
            for k in keys:
                self._pick_list.addItem(k)


class AtlasDisplay(ComponentTab):
    def __init__(self, extractor: DataAPI):
        super().__init__()
        self._extractor = extractor
        self.grid = QGridLayout()
        self.setLayout(self.grid)
        atlas_fig = Figure(tight_layout=True)
        atlas_fig.set_facecolor("gray")
        self._atlas_stats_fig = atlas_fig.add_subplot(111)
        self._atlas_stats_fig.set_facecolor("silver")
        self._atlas_stats = FigureCanvasQTAgg(atlas_fig)
        self._data: Dict[str, List[float]] = {}
        self._grid_square_averages: Dict[str, Dict[str, float]] = {}
        self._particle_data: Dict[str, List[float]] = {}
        self._colour_bar = None
        self._grid_square: Optional[GridSquare] = None
        self._all_grid_squares: List[GridSquare] = []
        self.project = ""

    def load(
        self,
        epu_dir: Path,
        grid_square: Optional[GridSquare] = None,
        all_grid_squares: Optional[List[GridSquare]] = None,
        data_changed: bool = False,
    ):
        self._grid_square = grid_square
        self._all_grid_squares = all_grid_squares or []
        if data_changed:
            self._update_atlas_stats()
        atlas_lbl = self._draw_atlas(
            epu_dir,
            grid_square=self._grid_square,
            all_grid_squares=self._all_grid_squares,
        )
        if atlas_lbl:
            vbox = QVBoxLayout()
            vbox.addWidget(atlas_lbl)
            vbox.addStretch()
            self.grid.addLayout(vbox, 0, 0)
            if self._grid_square:
                tile_lbl = self._draw_tile(self._grid_square, epu_dir)
                vbox = QVBoxLayout()
                vbox.addWidget(tile_lbl)
                vbox.addWidget(self._atlas_stats)
                vbox.addStretch()
                self.grid.addLayout(vbox, 0, 1)

    def _update_atlas_stats(self):
        atlas_fig = Figure(tight_layout=True)
        atlas_fig.set_facecolor("gray")
        self._atlas_stats_fig = atlas_fig.add_subplot(111)
        self._atlas_stats_fig.set_facecolor("silver")
        self._atlas_stats = InteractivePlot(atlas_fig)
        try:
            if self._colour_bar:
                self._colour_bar.remove()
        except (AttributeError, ValueError):
            pass
        if len(self._data.keys()) == 1:
            self._atlas_stats.set_data(list(self._data.values())[0])
            self._atlas_stats_fig.hist(
                list(self._data.values())[0], color="darkturquoise"
            )
            self._atlas_stats_fig.axes.set_xlabel(list(self._data.keys())[0])
        if len(self._data.keys()) == 2:
            labels = []
            data = []
            for k, v in self._data.items():
                labels.append(k)
                data.append(v)
            self._atlas_stats.set_data(data)
            self._atlas_stats_fig.scatter(data[0], data[1], color="darkturquoise")
            self._atlas_stats_fig.axes.set_xlabel(labels[0])
            self._atlas_stats_fig.axes.set_ylabel(labels[1])
        if len(self._data.keys()) > 2:
            labels = []
            data = []
            for k, v in self._data.items():
                labels.append(k)
                data.append(list(v))
            corr = np.corrcoef(data)
            mat = self._atlas_stats_fig.matshow(corr)
            ticks_loc = (
                self._atlas_stats_fig.axes.get_xticks(),
                self._atlas_stats_fig.axes.get_yticks(),
            )
            self._atlas_stats.set_data(corr)
            self._atlas_stats_fig.xaxis.set_major_locator(
                mticker.FixedLocator(ticks_loc[0][1:-1])
            )
            self._atlas_stats_fig.yaxis.set_major_locator(
                mticker.FixedLocator(ticks_loc[1][1:-1])
            )
            self._atlas_stats_fig.axes.set_xticklabels(labels, rotation=45)
            self._atlas_stats_fig.axes.set_yticklabels(labels)
            self._colour_bar = self._atlas_stats_fig.figure.colorbar(mat)
        self._atlas_stats.draw()

    def _draw_atlas(
        self,
        epu_dir: Path,
        grid_square: Optional[GridSquare] = None,
        all_grid_squares: Optional[List[GridSquare]] = None,
        flip: Tuple[int, int] = (1, 1),
    ) -> Optional[QLabel]:
        _atlas: Optional[Atlas] = None
        _atlases = self._extractor.get_atlases(project=self.project)
        if isinstance(_atlases, Atlas):
            _atlas = _atlases
        elif _atlases:
            _atlas = _atlases[0]
        if _atlas:
            atlas_pixmap = QPixmap(_atlas.thumbnail)
            if flip != (1, 1):
                atlas_pixmap = atlas_pixmap.transformed(QTransform().scale(*flip))
            if grid_square:
                imvs: Optional[list] = None
                if (
                    self._data
                    and grid_square
                    and all_grid_squares
                    and len(self._data.keys()) == 1
                ):
                    imvs = [
                        list(self._grid_square_averages.values())[0].get(
                            gs.grid_square_name
                        )
                        for gs in all_grid_squares
                        if gs != grid_square
                    ]
                    imvs = list(np.nan_to_num(imvs))
                qsize = atlas_pixmap.size()
                atlas_lbl = ImageLabel(
                    _atlas,
                    grid_square,
                    (qsize.width(), qsize.height()),
                    epu_dir,
                    parent=self,
                    overwrite_readout=True,
                    value=list(self._grid_square_averages.values())[0].get(
                        grid_square.grid_square_name
                    )
                    if imvs
                    else None,
                    extra_images=[gs for gs in all_grid_squares if gs != grid_square]
                    if all_grid_squares
                    else [],
                    image_values=imvs,
                )
                atlas_lbl.setPixmap(atlas_pixmap)
            else:
                atlas_lbl = QLabel(self)
                atlas_lbl.setPixmap(atlas_pixmap)
            return atlas_lbl
        return None

    def _draw_tile(
        self, grid_square: GridSquare, epu_dir: Path, flip: Tuple[int, int] = (1, 1)
    ) -> Optional[QLabel]:
        _tile = self._extractor.get_tile(
            (grid_square.stage_position_x, grid_square.stage_position_y),
            project=self.project,
        )
        if _tile:
            tile_pixmap = QPixmap(_tile.thumbnail)
            if flip != (1, 1):
                tile_pixmap = tile_pixmap.transformed(QTransform().scale(*flip))
            qsize = tile_pixmap.size()
            tile_lbl = ImageLabel(
                _tile,
                grid_square,
                (qsize.width(), qsize.height()),
                epu_dir,
                parent=self,
            )
            tile_lbl.setPixmap(tile_pixmap)
            return tile_lbl
        return None



import argparse
from pathlib import Path

from smartem.data_model.extract import DataAPI
from smartem.parsing.export import export_foil_holes


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-o",
        "--out_dir",
        help="Directory to export to",
        dest="out_dir",
        default=".",
    )
    parser.add_argument(
        "-p",
        "--projects",
        nargs="+",
        help="Names of smartEM projects to export",
        dest="projects",
    )
    parser.add_argument(
        "--use_adjusted_stage",
        action="store_true",
        dest="use_adjusted_stage",
    )
    args = parser.parse_args()

    data_api = DataAPI()
    export_foil_holes(
        data_api,
        out_dir=Path(args.out_dir),
        projects=args.projects,
        use_adjusted_stage=args.use_adjusted_stage,
    )



from multiprocessing import Process
from threading import Thread
from typing import List, Optional

from PyQt5.QtWidgets import QWidget


def background(
    children: Optional[List[str]] = None, lock_self: bool = False, process: bool = False
):
    if not children:
        children = []

    def background_decorator(func, *args, **kwargs):
        def wrapper(cl, *args, **kwargs):
            if cl._thread and cl._thread.is_alive():
                return
            frozen_widgets = [getattr(cl, ch) for ch in children]
            if lock_self:
                frozen_widgets.append(cl)
            if process:
                cl._thread = Process(
                    target=cl._background,
                    args=(func, *args),
                    kwargs={"_children": frozen_widgets, **kwargs},
                )
            else:
                cl._thread = Thread(
                    target=cl._background,
                    args=(func, *args),
                    kwargs={"_children": frozen_widgets, **kwargs},
                )
            cl._thread.start()

        return wrapper

    return background_decorator


class ComponentTab(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__()
        if kwargs.get("refreshers"):
            self._refreshers = kwargs["refreshers"]
        else:
            self._refreshers = []
        self._thread: Optional[Thread] = None

    def refresh(self):
        for refr in self._refreshers:
            refr.refresh()

    def _background(self, _background_process, *args, **kwargs):
        children = kwargs.get("_children")
        try:
            kwargs.pop("_children")
        except KeyError:
            pass
        if children is not None:
            for ch in children:
                ch.setEnabled(False)
        else:
            for ch in self.findChildren(QWidget):
                ch.setEnabled(False)
        _background_process(self, *args, **kwargs)
        if children is not None:
            for ch in children:
                ch.setEnabled(True)
        else:
            for ch in self.findChildren(QWidget):
                ch.setEnabled(True)



from __future__ import annotations

import importlib.resources

import matplotlib

matplotlib.use("Qt5Agg")
from pathlib import Path
from typing import List, Optional

from PyQt5.QtWidgets import (
    QApplication,
    QComboBox,
    QFileDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

import smartem.gui
from smartem.data_model import Project
from smartem.data_model.extract import DataAPI
from smartem.gui.qt.component_tab import ComponentTab
from smartem.gui.qt.display import AtlasDisplay, MainDisplay
from smartem.gui.qt.loader import (
    ExposureDataLoader,
    ParticleDataLoader,
    ParticleSetDataLoader,
)
from smartem.gui.qt.loader_csv import CSVDataLoader
from smartem.parsing.epu import create_atlas_and_tiles, parse_epu_dir, parse_epu_version
from smartem.parsing.relion_default import gather_relion_defaults


class App:
    def __init__(self, extractor: DataAPI):
        self.app = QApplication([])
        self.window = QtFrame(extractor)
        self.app.setStyleSheet(
            importlib.resources.read_text(smartem.gui.qt, "qt_style.css")
        )

    def start(self):
        self.window.resize(1600, 900)
        self.window.show()
        self.app.exec()


class QtFrame(QWidget):
    def __init__(self, extractor: DataAPI):
        super().__init__()
        self.tabs = QTabWidget()
        self.layout = QVBoxLayout(self)
        atlas_display = AtlasDisplay(extractor)
        main_display = MainDisplay(extractor, atlas_view=atlas_display)
        data_loader = ExposureDataLoader(extractor, refreshers=[main_display])
        csv_data_loader = CSVDataLoader(extractor, refreshers=[main_display])
        particle_loader = ParticleDataLoader(extractor, refreshers=[main_display])
        particle_set_loader = ParticleSetDataLoader(
            extractor, refreshers=[main_display]
        )
        proj_loader = ProjectLoader(
            extractor,
            data_loader,
            csv_data_loader,
            particle_loader,
            particle_set_loader,
            main_display,
            atlas_display,
        )
        self.tabs.addTab(proj_loader, "Project")
        self.tabs.addTab(data_loader, "Load mic data")
        self.tabs.addTab(csv_data_loader, "Load mic data (CSV)")
        self.tabs.addTab(particle_loader, "Load particle data")
        self.tabs.addTab(particle_set_loader, "Load particle set data")
        self.tabs.addTab(main_display, "Grid square view")
        self.tabs.addTab(atlas_display, "Atlas view")
        self.layout.addWidget(self.tabs)
        self.setLayout(self.layout)


class ProjectLoader(ComponentTab):
    def __init__(
        self,
        extractor: DataAPI,
        data_loader: ExposureDataLoader,
        csv_data_loader: CSVDataLoader,
        particle_loader: ParticleDataLoader,
        particle_set_loader: ParticleSetDataLoader,
        main_display: MainDisplay,
        atlas_display: AtlasDisplay,
        refreshers: Optional[List[ComponentTab]] = None,
    ):
        super().__init__(refreshers=refreshers)
        self._extractor = extractor
        self._data_loader = data_loader
        self._csv_data_loader = csv_data_loader
        self._particle_loader = particle_loader
        self._particle_set_loader = particle_set_loader
        self._main_display = main_display
        self._atlas_display = atlas_display
        self.grid = QGridLayout()
        self.setLayout(self.grid)
        self.epu_dir = ""
        self.project_dir = ""
        self._combo = QComboBox()
        projects = self._extractor.get_projects()
        self._combo.addItem(None)
        for proj in projects:
            self._combo.addItem(proj)
        self._combo.currentIndexChanged.connect(self._select_project)
        self._project_name = self._combo.currentText()
        self._name_input = QLineEdit()
        self._name_input.returnPressed.connect(self._button_check)
        hbox = QHBoxLayout()
        hbox.addWidget(self._combo, 1)
        hbox.addWidget(self._name_input, 1)
        self.grid.addLayout(hbox, 1, 1, 1, 2)
        epu_hbox = QHBoxLayout()
        epu_btn = QPushButton("Select EPU directory")
        epu_btn.clicked.connect(self._select_epu_dir)
        self.epu_lbl = QLabel()
        self.epu_lbl.setText(f"Selected: {self.epu_dir}")
        epu_hbox.addWidget(epu_btn, 1)
        epu_hbox.addWidget(self.epu_lbl, 1)
        self.grid.addLayout(epu_hbox, 2, 1, 1, 2)
        self.atlas = None
        atlas_hbox = QHBoxLayout()
        atlas_btn = QPushButton("Select Atlas")
        atlas_btn.clicked.connect(self._select_atlas)
        self.atlas_lbl = QLabel()
        self.atlas_lbl.setText(f"Selected: {self.atlas}")
        atlas_hbox.addWidget(atlas_btn, 1)
        atlas_hbox.addWidget(self.atlas_lbl, 1)
        self.grid.addLayout(atlas_hbox, 3, 1, 1, 2)
        project_hbox = QHBoxLayout()
        project_btn = QPushButton("Select project directory")
        project_btn.clicked.connect(self._select_processing_project)
        self.project_lbl = QLabel()
        self.project_lbl.setText(f"Selected: {self.project_dir}")
        project_hbox.addWidget(project_btn)
        project_hbox.addWidget(self.project_lbl)
        self.grid.addLayout(project_hbox, 4, 1, 1, 2)
        self._load_btn = QPushButton("Load")
        self._load_btn.clicked.connect(self.load)
        self.grid.addWidget(self._load_btn, 5, 1)
        self._create_btn = QPushButton("Create")
        self._create_btn.clicked.connect(self._create_project)
        self.grid.addWidget(self._create_btn, 5, 2)
        self._create_gather_btn = QPushButton("Create and load default data")
        self._create_gather_btn.clicked.connect(self._create_and_gather)
        self.grid.addWidget(self._create_gather_btn, 6, 2)
        self._button_check()

    def _button_check(self):
        if self._combo.currentText():
            self._load_btn.setEnabled(True)
            self._create_btn.setEnabled(False)
            self._create_gather_btn.setEnabled(False)
        elif self._name_input.text() and self.atlas and self.epu_dir:
            self._load_btn.setEnabled(False)
            self._create_btn.setEnabled(True)
            self._create_gather_btn.setEnabled(True)
        else:
            self._load_btn.setEnabled(False)
            self._create_btn.setEnabled(False)
            self._create_gather_btn.setEnabled(False)

    def _select_project(self):
        if self._combo.currentText():
            self._project_name = self._combo.currentText()
            self._name_input.setEnabled(False)
            project = self._extractor.get_project(project_name=self._project_name)
            atlas = self._extractor.get_atlas_from_project(project)
            self.epu_dir = project.acquisition_directory
            self.epu_lbl.setText(f"Selected: {self.epu_dir}")
            self.project_dir = project.processing_directory
            self.project_lbl.setText(f"Selected: {self.project_dir}")
            self.atlas = atlas.thumbnail
            self.atlas_lbl.setText(f"Selected: {self.atlas}")
        else:
            self._name_input.setEnabled(False)
            self._project_name = self._name_input.text()
        self._button_check()

    def _select_epu_dir(self):
        self.epu_dir = QFileDialog.getExistingDirectory(
            self, "Select EPU directory", ".", QFileDialog.ShowDirsOnly
        )
        self.epu_lbl.setText(f"Selected: {self.epu_dir}")
        self._button_check()

    def _select_atlas_combo(self, index: int):
        self.atlas = self._combo.currentText()
        self.atlas_lbl.setText(f"Selected: {self.atlas}")
        self._button_check()

    def _select_atlas(self):
        self.atlas = QFileDialog.getOpenFileName(self, "Select Atlas image", ".")[0]
        self.atlas_lbl.setText(f"Selected: {self.atlas}")
        self._button_check()

    def _select_processing_project(self):
        self.project_dir = QFileDialog.getExistingDirectory(
            self, "Select project directory", ".", QFileDialog.ShowDirsOnly
        )
        self.project_lbl.setText(f"Selected: {self.project_dir}")
        self._button_check()

    def _update_loaders(self):
        self._data_loader._set_project_directory(Path(self.project_dir))
        self._csv_data_loader._set_project_directory(Path(self.project_dir))
        self._particle_loader._set_project_directory(Path(self.project_dir))
        self._particle_set_loader._set_project_directory(Path(self.project_dir))
        self._data_loader.project = self._project_name
        self._csv_data_loader.project = self._project_name
        self._particle_loader.project = self._project_name
        self._particle_set_loader.project = self._project_name

    def _create_project(self):
        self._project_name = self._name_input.text()
        found = self._extractor.set_project(self._project_name)
        if not found:
            _atlas_id = create_atlas_and_tiles(Path(self.atlas), self._extractor)
            software, version = parse_epu_version(Path(self.epu_dir))
            proj = Project(
                atlas_id=_atlas_id,
                acquisition_directory=self.epu_dir,
                project_name=self._project_name,
                processing_directory=self.project_dir,
                acquisition_software=software,
                acquisition_software_version=version,
            )
            self._extractor.put([proj])
        atlas_found = self._extractor.set_project(self._project_name)
        if not atlas_found:
            raise ValueError(
                "Project record not found despite having just been inserted"
            )
        parse_epu_dir(Path(self.epu_dir), self._extractor, self._project_name)
        self._main_display._set_epu_directory(Path(self.epu_dir))
        self._main_display._set_data_size(Path(self.project_dir))
        self._main_display.set_project(self._project_name)
        self._atlas_display.project = self._project_name
        self.refresh()
        self._update_loaders()

    def load(self):
        atlas_found = self._extractor.set_project(self._project_name)
        if not atlas_found:
            raise ValueError("Atlas record not found")
        self._main_display._set_epu_directory(Path(self.epu_dir))
        self._main_display._set_data_size(Path(self.project_dir))
        self._main_display.set_project(self._project_name)
        self._atlas_display.project = self._project_name
        self.refresh()
        self._update_loaders()

    def _create_and_gather(self):
        self._create_project()
        gather_relion_defaults(
            Path(self.project_dir), self._extractor, self._project_name
        )
        self.refresh()

    def refresh(self):
        super().refresh()
        self._main_display.load()
        self._atlas_display.load(Path(self.epu_dir))



import math
from itertools import cycle
from pathlib import Path
from typing import List, Optional, Tuple, Union

import matplotlib
import mrcfile
import numpy as np
from PyQt5 import QtCore
from PyQt5.QtGui import QBrush, QColor, QPainter, QPen
from PyQt5.QtWidgets import QComboBox, QLabel

from smartem.data_model import Atlas, Exposure, FoilHole, GridSquare, Particle, Tile
from smartem.stage_model import StageCalibration, find_point_pixel


def colour_gradient(value: float) -> str:
    low = "#EF3054"
    high = "#47682C"
    low_rgb = np.array(matplotlib.colors.to_rgb(low))
    high_rgb = np.array(matplotlib.colors.to_rgb(high))
    return matplotlib.colors.to_hex((1 - value) * low_rgb + value * high_rgb)


class ParticleImageLabel(QLabel):
    def __init__(
        self,
        image: Exposure,
        particles: Union[List[Particle], List[List[Particle]]],
        image_size: Tuple[int, int],
        image_scale: float = 0.5,
        selection_box: Optional[QComboBox] = None,
        particle_diameter: int = 30,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._image = image
        self._image_size = image_size
        self._particles = particles
        self._image_scale = image_scale
        self._selection_box = selection_box
        self._diameter = particle_diameter

    def mousePressEvent(self, ev):
        if self._selection_box is not None:
            self._selection_box.setFocus()
            self._selection_box.activateWindow()

    def draw_circle(
        self, coordinates: Tuple[float, float], diameter: int, painter: QPainter
    ):
        if self._image.readout_area_x and self._image.readout_area_y:
            x = (
                coordinates[0]
                * self._image_scale
                * (self._image_size[0] / self._image.readout_area_x)
                - diameter / 2
            )
            y = (
                coordinates[1]
                * self._image_scale
                * (self._image_size[1] / self._image.readout_area_y)
                - diameter / 2
            )
            painter.drawEllipse(int(x), int(y), diameter, diameter)

    def paintEvent(self, e):
        super().paintEvent(e)

        painter = QPainter(self)
        pen = QPen(QColor(QtCore.Qt.blue))
        pen.setWidth(3)
        painter.setPen(pen)

        colour_cycle = cycle(
            [
                QColor(QtCore.Qt.red),
                QColor(QtCore.Qt.green),
                QColor(QtCore.Qt.cyan),
                QColor(QtCore.Qt.blue),
            ]
        )

        if self._particles and isinstance(self._particles[0], Particle):
            for particle in self._particles:
                self.draw_circle((particle.x, particle.y), self._diameter, painter)
        elif self._particles:
            for particle_group in self._particles:
                pen = QPen(next(colour_cycle))
                pen.setWidth(3)
                painter.setPen(pen)
                for particle in particle_group:
                    self.draw_circle((particle.x, particle.y), self._diameter, painter)

        painter.end()


class ImageLabel(QLabel):
    def __init__(
        self,
        image: Union[Atlas, Tile, GridSquare, FoilHole, Exposure],
        contained_image: Optional[Union[GridSquare, FoilHole, Exposure]],
        image_size: Tuple[int, int],
        image_directory: Path,
        overwrite_readout: bool = False,
        value: Optional[float] = None,
        extra_images: Optional[list] = None,
        image_values: Optional[List[float]] = None,
        selection_box: Optional[QComboBox] = None,
        stage_calibration: Optional[StageCalibration] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._image = image
        self._image_directory = image_directory
        self._contained_image = contained_image
        self._extra_images = extra_images or []
        self._image_size = image_size
        self._overwrite_readout = overwrite_readout
        self._value = value
        self._image_values = image_values or []
        self._selection_box = selection_box
        self._stage_calibration = stage_calibration or StageCalibration()

    def mousePressEvent(self, ev):
        if self._selection_box is not None:
            self._selection_box.setFocus()
            self._selection_box.activateWindow()

    def draw_rectangle(
        self,
        inner_image: Union[GridSquare, FoilHole, Exposure],
        readout_area: Tuple[int, int],
        scaled_pixel_size: float,
        painter: QPainter,
        normalised_value: Optional[float] = None,
    ):
        if normalised_value is not None:
            c = QColor()
            rgb = (
                int(255 * x)
                for x in matplotlib.colors.to_rgb(colour_gradient(normalised_value))
            )
            c.setRgb(*rgb, alpha=150)
            brush = QBrush(c, QtCore.Qt.SolidPattern)
            painter.setBrush(brush)
        else:
            brush = QBrush()
            painter.setBrush(brush)
        if inner_image.thumbnail or hasattr(inner_image, "adjusted_stage_position_x"):
            if (
                isinstance(inner_image, FoilHole)
                and inner_image.adjusted_stage_position_x is not None
            ):
                rect_centre = find_point_pixel(
                    (
                        inner_image.adjusted_stage_position_x,
                        inner_image.adjusted_stage_position_y,
                    ),
                    (self._image.stage_position_x, self._image.stage_position_y),
                    scaled_pixel_size,
                    (
                        int(
                            readout_area[0]
                            / (scaled_pixel_size / self._image.pixel_size)
                        ),
                        int(
                            readout_area[1]
                            / (scaled_pixel_size / self._image.pixel_size)
                        ),
                    ),
                    xfactor=-1 if self._stage_calibration.x_flip else 1,
                    yfactor=-1 if self._stage_calibration.y_flip else 1,
                )
            else:
                rect_centre = find_point_pixel(
                    (
                        inner_image.stage_position_x,
                        inner_image.stage_position_y,
                    ),
                    (self._image.stage_position_x, self._image.stage_position_y),
                    scaled_pixel_size,
                    (
                        int(
                            readout_area[0]
                            / (scaled_pixel_size / self._image.pixel_size)
                        ),
                        int(
                            readout_area[1]
                            / (scaled_pixel_size / self._image.pixel_size)
                        ),
                    ),
                    xfactor=-1 if self._stage_calibration.x_flip else 1,
                    yfactor=-1 if self._stage_calibration.y_flip else 1,
                )
            if not inner_image.thumbnail:
                edge_lengths = (10, 10)
                painter.drawEllipse(
                    int(rect_centre[1] - 0.5 * edge_lengths[0])
                    if self._stage_calibration.inverted
                    else int(rect_centre[0] - 0.5 * edge_lengths[0]),
                    int(rect_centre[0] - 0.5 * edge_lengths[1])
                    if self._stage_calibration.inverted
                    else int(rect_centre[1] - 0.5 * edge_lengths[1]),
                    edge_lengths[0],
                    edge_lengths[1],
                )
            else:
                edge_lengths = (
                    int(
                        inner_image.readout_area_x
                        * inner_image.pixel_size
                        / scaled_pixel_size
                    ),
                    int(
                        inner_image.readout_area_y
                        * inner_image.pixel_size
                        / scaled_pixel_size
                    ),
                )
                painter.drawRect(
                    int(rect_centre[1] - 0.5 * edge_lengths[0])
                    if self._stage_calibration.inverted
                    else int(rect_centre[0] - 0.5 * edge_lengths[0]),
                    int(rect_centre[0] - 0.5 * edge_lengths[1])
                    if self._stage_calibration.inverted
                    else int(rect_centre[1] - 0.5 * edge_lengths[1]),
                    edge_lengths[0],
                    edge_lengths[1],
                )

    def paintEvent(self, e):
        super().paintEvent(e)

        if self._contained_image:
            painter = QPainter(self)
            pen = QPen(QColor(QtCore.Qt.blue))
            pen.setWidth(3)
            painter.setPen(pen)
            if self._overwrite_readout:
                with mrcfile.open(
                    (self._image_directory / self._image.thumbnail).with_suffix(".mrc")
                ) as mrc:
                    readout_area = mrc.data.shape
            else:
                readout_area = (self._image.readout_area_x, self._image.readout_area_y)
            scaled_pixel_size = self._image.pixel_size * (
                readout_area[0] / self._image_size[0]
            )

            if self._image_values:
                try:
                    min_value = min(
                        imv
                        for imv in self._image_values + [self._value]
                        if imv is not None
                        and not math.isnan(imv)
                        and not math.isinf(imv)
                    )
                # catch when an empty sequence is passed to min
                except ValueError:
                    return
                shifted = [
                    iv - min_value
                    if iv is not None and not math.isnan(iv) and not math.isinf(iv)
                    else None
                    for iv in self._image_values
                ]
                all_shifted = [
                    iv - min_value
                    if iv is not None and not math.isnan(iv) and not math.isinf(iv)
                    else None
                    for iv in self._image_values + [self._value]
                ]
                maxv = max(
                    abs(imv)
                    for imv in all_shifted
                    if imv is not None and not math.isnan(imv) and not math.isinf(imv)
                )
                if maxv:
                    normalised = [
                        s / maxv
                        if s is not None and not math.isnan(s) and not math.isinf(s)
                        else None
                        for s in shifted
                    ]
                else:
                    normalised = shifted
            for i, im in enumerate(self._extra_images):
                if self._image_values:
                    self.draw_rectangle(
                        im,
                        readout_area,
                        scaled_pixel_size,
                        painter,
                        normalised_value=normalised[i],
                    )
                else:
                    self.draw_rectangle(im, readout_area, scaled_pixel_size, painter)

            pen = QPen(QColor(QtCore.Qt.red))
            pen.setWidth(3)
            painter.setPen(pen)

            if self._value is not None and self._image_values:
                norm_value = (self._value - min_value) / maxv
            else:
                norm_value = 0

            norm_value = np.nan_to_num(norm_value)

            if self._value is not None:
                self.draw_rectangle(
                    self._contained_image,
                    readout_area,
                    scaled_pixel_size,
                    painter,
                    normalised_value=norm_value,
                )
            else:
                self.draw_rectangle(
                    self._contained_image, readout_area, scaled_pixel_size, painter
                )

            painter.end()



import json
import os
from pathlib import Path
from typing import Dict, List, Optional

from gemmi import cif

from smartem.data_model import (
    ExposureInfo,
    Particle,
    ParticleInfo,
    ParticleSet,
    ParticleSetInfo,
    ParticleSetLinker,
)
from smartem.data_model.extract import DataAPI


def open_star_file(star_file_path: Path):
    gemmi_readable_path = os.fspath(star_file_path)
    return cif.read_file(gemmi_readable_path)


def get_columns(star_file, ignore: Optional[List[str]] = None) -> List[str]:
    json_star = json.loads(star_file.as_json())
    cols = []
    for v in json_star.values():
        if ignore:
            vals = [_v for _v in v.keys() if all(ig not in _v for ig in ignore)]
            cols.extend(vals)
        else:
            cols.extend(v.keys())
    return cols


def get_column_data(
    star_file, columns: List[str], block_tag: str
) -> Dict[str, List[str]]:
    json_star = json.loads(star_file.as_json())
    return {k: v for k, v in json_star[block_tag].items() if k in columns}


def insert_exposure_data(
    data: Dict[str, List[str]],
    exposure_tag: str,
    star_file_path: str,
    extractor: DataAPI,
    validate: bool = True,
    extra_suffix: str = "",
    project: str = "",
):
    if validate:
        exposures = [e.exposure_name for e in extractor.get_exposures(project=project)]
    exposure_info: List[ExposureInfo] = []
    for k, v in data.items():
        if k != exposure_tag:
            for i, value in enumerate(v):
                exposure_name = (
                    Path(data[exposure_tag][i])
                    .stem.replace("_fractions", "")
                    .replace("_Fractions", "")
                    .replace(extra_suffix, "")
                    + ".jpg"
                )
                if validate:
                    if exposure_name in exposures:
                        exinf = ExposureInfo(
                            exposure_name=exposure_name,
                            source=star_file_path,
                            key=k,
                            value=value,
                        )
                        exposure_info.append(exinf)
                    else:
                        pass
                        # print(f"exposure {exposure_name} not found")
                else:
                    exinf = ExposureInfo(
                        exposure_name=exposure_name,
                        source=star_file_path,
                        key=k,
                        value=value,
                    )
                    exposure_info.append(exinf)
    extractor.put(exposure_info)


def _structure_particle_data(
    data: Dict[str, List[str]],
    exposures: List[str],
    exposure_tag: str,
    x_tag: str,
    y_tag: str,
) -> Dict[str, Dict[str, list]]:
    structured_data: Dict[str, Dict[str, list]] = {}
    for i, micrograph_path in enumerate(data[exposure_tag]):
        exposure_name = (
            Path(micrograph_path)
            .stem.replace("_fractions", "")
            .replace("_Fractions", "")
            + ".jpg"
        )
        if exposure_name in exposures:
            try:
                structured_data[exposure_name]["coordinates"].append(
                    (data[x_tag][i], data[y_tag][i])
                )
                structured_data[exposure_name]["indices"].append(i)
            except KeyError:
                structured_data[exposure_name] = {}
                structured_data[exposure_name]["coordinates"] = [
                    (data[x_tag][i], data[y_tag][i])
                ]
                structured_data[exposure_name]["indices"] = [i]
    return structured_data


def insert_particle_data(
    data: Dict[str, List[str]],
    exposure_tag: str,
    x_tag: str,
    y_tag: str,
    star_file_path: str,
    extractor: DataAPI,
    project: str = "",
):
    exposures = [e.exposure_name for e in extractor.get_exposures(project=project)]
    particle_info: List[ParticleInfo] = []
    extra_keys = [k for k in data.keys() if k and k not in (exposure_tag, x_tag, y_tag)]

    structured_data = _structure_particle_data(
        data, exposures, exposure_tag, x_tag, y_tag
    )

    new_particles = []
    particle_info = []
    for exposure in exposures:
        if structured_data.get(exposure):
            existing_particles = extractor.get_particles(exposure_name=exposure)
            existing_particle_coords = {
                (ep.x, ep.y): ep.particle_id for ep in existing_particles
            }
            for i, particle in enumerate(structured_data[exposure]["coordinates"]):
                if existing_particle_coords.get(particle):
                    for k in extra_keys:
                        particle_info.append(
                            ParticleInfo(
                                particle_id=existing_particle_coords[particle],
                                source=star_file_path,
                                key=k,
                                value=data[k][structured_data[exposure]["indices"][i]],
                            )
                        )
                else:
                    new_particles.append(
                        Particle(x=particle[0], y=particle[1], exposure_name=exposure)
                    )

    if new_particles:
        new_particle_ids = extractor.put(new_particles)
        for exposure in exposures:
            if structured_data.get(exposure):
                existing_particle_coords = {
                    (p.x, p.y): pid.particle_id
                    for p, pid in zip(new_particles, new_particle_ids)
                    if p.exposure_name == exposure
                }
                for i, particle in enumerate(structured_data[exposure]["coordinates"]):
                    if existing_particle_coords.get(particle):
                        for k in extra_keys:
                            particle_info.append(
                                ParticleInfo(
                                    particle_id=existing_particle_coords[particle],
                                    source=star_file_path,
                                    key=k,
                                    value=data[k][
                                        structured_data[exposure]["indices"][i]
                                    ],
                                )
                            )

    if particle_info:
        extractor.put(particle_info)


def insert_particle_set(
    data: Dict[str, List[str]],
    set_name: str,
    set_id_tag: str,
    exposure_tag: str,
    x_tag: str,
    y_tag: str,
    star_file_path: str,
    extractor: DataAPI,
    project: str,
    add_source_to_id: bool = False,
):
    extra_keys = [
        k
        for k in data.keys()
        if k and k not in (exposure_tag, x_tag, y_tag, set_id_tag)
    ]
    set_ids = set(data[set_id_tag])
    _particle_sets = extractor.get_particle_sets(
        project, set_name, set_ids, star_file_path
    )
    if not _particle_sets:
        if add_source_to_id:
            particle_sets = [
                ParticleSet(
                    group_name=set_name,
                    identifier=star_file_path + ":" + str(set_id),
                    project_name=project,
                )
                for set_id in set_ids
            ]
        else:
            particle_sets = [
                ParticleSet(
                    group_name=set_name,
                    identifier=str(set_id),
                    project_name=project,
                )
                for set_id in set_ids
            ]
        extractor.put(particle_sets)
    else:
        particle_sets = []
        for set_id in set_ids:
            for ps in _particle_sets:
                if ps.identifier == set_id:
                    particle_sets.append(ps)
    exposures = [e.exposure_name for e in extractor.get_exposures(project=project)]
    structured_data = _structure_particle_data(
        data, exposures, exposure_tag, x_tag, y_tag
    )
    set_instances: Dict[str, Dict[str, float]] = {}
    linkers = []
    new_particles = []
    linkers_for_new_particles = []

    if extra_keys:
        for si in set_ids:
            for i, s in enumerate(data[set_id_tag]):
                if s == si:
                    set_instances[si] = {
                        k: float(v[i]) for k, v in data.items() if k in extra_keys
                    }
                    break

    for exposure in exposures:
        if structured_data.get(exposure):
            particles = extractor.get_particles(exposure_name=exposure)
            particle_coords = {(p.x, p.y): p.particle_id for p in particles}
            for i, particle in enumerate(structured_data[exposure]["coordinates"]):
                if particle_coords.get(particle):
                    if not _particle_sets:
                        if add_source_to_id:
                            linkers.append(
                                ParticleSetLinker(
                                    set_name=star_file_path
                                    + ":"
                                    + str(
                                        data[set_id_tag][
                                            structured_data[exposure]["indices"][i]
                                        ]
                                    ),
                                    particle_id=particle_coords[particle],
                                )
                            )
                        else:
                            linkers.append(
                                ParticleSetLinker(
                                    set_name=data[set_id_tag][
                                        structured_data[exposure]["indices"][i]
                                    ],
                                    particle_id=particle_coords[particle],
                                )
                            )
                else:
                    new_particles.append(
                        Particle(x=particle[0], y=particle[1], exposure_name=exposure)
                    )

    if linkers:
        extractor.put(linkers)

    if new_particles:
        new_particle_ids = extractor.put(new_particles)

        for exposure in exposures:
            if structured_data.get(exposure):
                particle_coords = {
                    (p.x, p.y): pid.particle_id
                    for p, pid in zip(new_particles, new_particle_ids)
                    if p.exposure_name == exposure
                }
                for i, particle in enumerate(structured_data[exposure]["coordinates"]):
                    if particle_coords.get(particle):
                        if add_source_to_id:
                            linkers_for_new_particles.append(
                                ParticleSetLinker(
                                    set_name=star_file_path
                                    + ":"
                                    + str(
                                        data[set_id_tag][
                                            structured_data[exposure]["indices"][i]
                                        ]
                                    ),
                                    particle_id=particle_coords[particle],
                                )
                            )
                        else:
                            linkers_for_new_particles.append(
                                ParticleSetLinker(
                                    set_name=data[set_id_tag][
                                        structured_data[exposure]["indices"][i]
                                    ],
                                    particle_id=particle_coords[particle],
                                )
                            )
        if linkers_for_new_particles:
            extractor.put(linkers_for_new_particles)

    particle_set_info = []
    for k in extra_keys:
        for si in set_ids:
            if add_source_to_id:
                particle_set_info.append(
                    ParticleSetInfo(
                        set_name=star_file_path + ":" + str(si),
                        source=star_file_path,
                        key=k,
                        value=set_instances[si][k],
                    )
                )
            else:
                particle_set_info.append(
                    ParticleSetInfo(
                        set_name=si,
                        source=star_file_path,
                        key=k,
                        value=set_instances[si][k],
                    )
                )
    if particle_set_info:
        extractor.put(particle_set_info)



import shutil
from pathlib import Path
from typing import Dict, List, Optional

import yaml
from pandas import DataFrame

from smartem.data_model.extract import DataAPI
from smartem.data_model.structure import extract_keys_with_foil_hole_averages
from smartem.parsing.epu import calibrate_coordinate_system


def export_foil_holes(
    data_api: DataAPI,
    out_dir: Path = Path("."),
    projects: Optional[List[str]] = None,
    use_adjusted_stage: bool = False,
):
    if not projects:
        projects = [data_api._project]
    out_gs_paths = {}
    data: Dict[str, list] = {
        "grid_square": [],
        "grid_square_pixel_size": [],
        "grid_square_x": [],
        "grid_square_y": [],
        "foil_hole": [],
        "foil_hole_pixel_size": [],
        "foil_hole_x": [],
        "foil_hole_y": [],
        "accummotiontotal": [],
        "ctfmaxresolution": [],
        "estimatedresolution": [],
        "maxvalueprobdistribution": [],
    }

    for project in projects:
        grid_squares = data_api.get_grid_squares(project)
        foil_holes = data_api.get_foil_holes(project=project)

        data_labels = [
            "_rlnaccummotiontotal",
            "_rlnctfmaxresolution",
            "_rlnestimatedresolution",
            "_rlnmaxvalueprobdistribution",
        ]

        _project = data_api.get_project(project_name=project)
        atlas = data_api.get_atlas_from_project(_project)
        atlas_id = atlas.atlas_id
        atlas_info = data_api.get_atlas_info(
            atlas_id,
            ["_rlnaccummotiontotal", "_rlnctfmaxresolution"],
            ["_rlnmaxvalueprobdistribution"],
            ["_rlnestimatedresolution"],
        )

        fh_extracted = extract_keys_with_foil_hole_averages(
            atlas_info,
            ["_rlnaccummotiontotal", "_rlnctfmaxresolution"],
            ["_rlnmaxvalueprobdistribution"],
            ["_rlnestimatedresolution"],
        )

        epu_dir = Path(_project.acquisition_directory)

        if not atlas.thumbnail:
            raise ValueError(f"No atlas image was found for {project}")
        atlas_image_path = Path(atlas.thumbnail)
        shutil.copy(atlas_image_path, out_dir / atlas_image_path.name)
        shutil.copy(
            atlas_image_path, out_dir / atlas_image_path.with_suffix(".mrc").name
        )

        gs_coordinates = {}
        gs_pixel_sizes = {}

        for gs in grid_squares:
            if gs.thumbnail:
                gs_dir = out_dir / gs.grid_square_name
                gs_dir.mkdir()
                thumbnail_path: Optional[Path] = epu_dir / gs.thumbnail
                if thumbnail_path:
                    shutil.copy(thumbnail_path, gs_dir / thumbnail_path.name)
                    shutil.copy(
                        thumbnail_path.with_suffix(".mrc"),
                        gs_dir / thumbnail_path.with_suffix(".mrc").name,
                    )
                    out_gs_paths[gs.grid_square_name] = (
                        gs_dir / thumbnail_path.name
                    ).relative_to(out_dir)
                gs_coordinates[gs.grid_square_name] = (
                    gs.stage_position_x,
                    gs.stage_position_y,
                )
                gs_pixel_sizes[gs.grid_square_name] = gs.pixel_size
        for fh in foil_holes:
            if all(
                fh_extracted[dl].averages is not None for dl in data_labels
            ):  # mypy doesn't accept this as good enough for the below
                if (
                    all(
                        fh_extracted[dl].averages.get(fh.foil_hole_name)  # type: ignore
                        for dl in data_labels
                    )
                    # and fh.thumbnail
                ):
                    thumbnail_path = None
                    if fh.thumbnail:
                        fh_dir = out_dir / fh.grid_square_name / fh.foil_hole_name
                        fh_dir.mkdir()
                        thumbnail_path = epu_dir / fh.thumbnail
                        shutil.copy(thumbnail_path, fh_dir / thumbnail_path.name)
                        shutil.copy(
                            thumbnail_path.with_suffix(".mrc"),
                            fh_dir / thumbnail_path.with_suffix(".mrc").name,
                        )
                    data["grid_square"].append(str(out_gs_paths[fh.grid_square_name]))
                    data["grid_square_pixel_size"].append(
                        gs_pixel_sizes[fh.grid_square_name]
                    )
                    data["grid_square_x"].append(gs_coordinates[fh.grid_square_name][0])
                    data["grid_square_y"].append(gs_coordinates[fh.grid_square_name][1])
                    data["foil_hole"].append(
                        str((fh_dir / thumbnail_path.name).relative_to(out_dir))
                        if thumbnail_path
                        else None
                    )
                    data["foil_hole_pixel_size"].append(fh.pixel_size)
                    if use_adjusted_stage:
                        data["foil_hole_x"].append(
                            fh.stage_position_x
                            if fh.adjusted_stage_position_x is None
                            else fh.adjusted_stage_position_x
                        )
                        data["foil_hole_y"].append(
                            fh.stage_position_y
                            if fh.adjusted_stage_position_y is None
                            else fh.adjusted_stage_position_y
                        )
                    else:
                        data["foil_hole_x"].append(fh.stage_position_x)
                        data["foil_hole_y"].append(fh.stage_position_y)
                    data["accummotiontotal"].append(
                        fh_extracted["_rlnaccummotiontotal"].averages[fh.foil_hole_name]  # type: ignore
                    )
                    data["ctfmaxresolution"].append(
                        fh_extracted["_rlnctfmaxresolution"].averages[fh.foil_hole_name]  # type: ignore
                    )
                    data["estimatedresolution"].append(
                        fh_extracted["_rlnestimatedresolution"].averages[  # type: ignore
                            fh.foil_hole_name
                        ]
                    )
                    data["maxvalueprobdistribution"].append(
                        fh_extracted["_rlnmaxvalueprobdistribution"].averages[  # type: ignore
                            fh.foil_hole_name
                        ]
                    )

    df = DataFrame.from_dict(data)
    df.to_csv(out_dir / "labels.csv", index=False)

    if projects:
        _project = data_api.get_project(project_name=projects[0])
        for dm in (Path(_project.acquisition_directory).parent / "Metadata").glob(
            "*.dm"
        ):
            stage_calibration = calibrate_coordinate_system(dm)
            if stage_calibration:
                with open(out_dir / "coordinate_calibration.yaml", "w") as outfile:
                    yaml.dump(stage_calibration._asdict(), outfile)
                break
